import spacy
import inflect
import itertools
import json
import pickle
from typing import Callable, Dict, Iterable, List
import uuid
import math

import numpy as np
from rouge_score import rouge_scorer, scoring
from sacrebleu import corpus_bleu
from torch import nn
from torch.utils.data import Sampler

from transformers import BartTokenizer

IGNORE_WORDS = set(["personx", "persony", "personz", "_", "'", "-"])
ROUGE_KEYS = ["rouge1", "rouge2", "rougeL"]


def vp_present_participle(phrase):
    nlp = spacy.load("en_core_web_sm")
    doc = nlp(phrase)
    inflection_engine = inflect.engine()
    return " ".join(
        [
            inflection_engine.present_participle(token.text)
            if token.pos_ == "VERB" and token.tag_ != "VGG"
            else token.text
            for token in doc
        ]
    )


def posessive(word):
    inflection_engine = inflect.engine()
    if inflection_engine.singular_noun(word) is False:
        return "have"
    else:
        return "has"


def article(word):
    return "an" if word[0] in ["a", "e", "i", "o", "u"] else "a"


def find_nth(haystack, needle, n):
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        start = haystack.find(needle, start + len(needle))
        n -= 1
    return start if start != -1 else None


def encode_line(
    tokenizer, line, max_length, pad_to_max_length=True, return_tensors="pt"
):
    extra_kw = (
        {"add_prefix_space": True} if isinstance(tokenizer, BartTokenizer) else {}
    )
    return tokenizer(
        [line],
        max_length=max_length,
        padding="max_length" if pad_to_max_length else None,
        truncation=True,
        return_tensors=return_tensors,
        **extra_kw,
    )


def lmap(f: Callable, x: Iterable) -> List:
    """list(map(f, x))"""
    return list(map(f, x))


def calculate_bleu_score(output_lns, refs_lns, **kwargs) -> dict:
    """Uses sacrebleu's corpus_bleu implementation."""
    return {"bleu": corpus_bleu(output_lns, [refs_lns], **kwargs).score}


def trim_batch(
    input_ids,
    pad_token_id,
    attention_mask=None,
):
    """Remove columns that are populated exclusively by pad_token_id"""
    keep_column_mask = input_ids.ne(pad_token_id).any(dim=0)
    if attention_mask is None:
        return input_ids[:, keep_column_mask]
    else:
        return (input_ids[:, keep_column_mask], attention_mask[:, keep_column_mask])


class SortishSampler(Sampler):
    "Go through the text data by order of src length with a bit of randomness. From fastai repo."

    def __init__(self, data, batch_size):
        self.data, self.bs = data, batch_size

    def key(self, i):
        return self.data[i]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self):
        idxs = np.random.permutation(len(self.data))
        sz = self.bs * 50
        ck_idx = [idxs[i : i + sz] for i in range(0, len(idxs), sz)]
        sort_idx = np.concatenate(
            [sorted(s, key=self.key, reverse=True) for s in ck_idx]
        )
        sz = self.bs
        ck_idx = [sort_idx[i : i + sz] for i in range(0, len(sort_idx), sz)]
        max_ck = np.argmax(
            [self.key(ck[0]) for ck in ck_idx]
        )  # find the chunk with the largest key,
        ck_idx[0], ck_idx[max_ck] = (
            ck_idx[max_ck],
            ck_idx[0],
        )  # then make sure it goes first.
        if len(ck_idx) < 3:
            return iter(np.concatenate(ck_idx))

        sort_idx = np.concatenate(
            [np.concatenate(np.random.permutation(ck_idx[1:-1])), ck_idx[-1]]
        )
        sort_idx = np.concatenate((ck_idx[0], sort_idx))
        return iter(sort_idx)


def pickle_load(path):
    """pickle.load(path)"""
    with open(path, "rb") as f:
        return pickle.load(f)


def pickle_save(obj, path):
    """pickle.dump(obj, path)"""
    with open(path, "wb") as f:
        return pickle.dump(obj, f)


def flatten_list(summary_ids: List[List]):
    return [x for x in itertools.chain.from_iterable(summary_ids)]


def save_json(content, path):
    with open(path, "w") as f:
        json.dump(content, f, indent=4)


def load_json(path):
    with open(path) as f:
        return json.load(f)


def calculate_rouge(
    output_lns: List[str], reference_lns: List[str], use_stemmer=True
) -> Dict:
    scorer = rouge_scorer.RougeScorer(ROUGE_KEYS, use_stemmer=use_stemmer)
    aggregator = scoring.BootstrapAggregator()

    for reference_ln, output_ln in zip(reference_lns, output_lns):
        scores = scorer.score(reference_ln, output_ln)
        aggregator.add_scores(scores)

    result = aggregator.aggregate()
    return {k: v.mid.fmeasure for k, v in result.items()}


def freeze_params(model: nn.Module):
    for par in model.parameters():
        par.requires_grad = False


def grad_status(model: nn.Module) -> Iterable:
    return (par.requires_grad for par in model.parameters())


def any_requires_grad(model: nn.Module) -> bool:
    return any(grad_status(model))


def assert_all_frozen(model):
    model_grads: List[bool] = list(grad_status(model))
    n_require_grad = sum(lmap(int, model_grads))
    npars = len(model_grads)
    assert not any(
        model_grads
    ), f"{n_require_grad/npars:.1%} of {npars} weights require grad"


def assert_not_all_frozen(model):
    model_grads: List[bool] = list(grad_status(model))
    npars = len(model_grads)
    assert any(model_grads), f"none of {npars} weights require grad"


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def get_uuid(length=8):
    u = str(uuid.uuid4())
    if length is not None:
        u = u[:length]
    return u


def truncate_sequences_dual(sequences, max_length):
    words_to_cut = sum(list(map(len, sequences))) - max_length
    if words_to_cut <= 0:
        return sequences

    words_to_cut_before = math.ceil(words_to_cut / 2.0)
    words_to_cut_after = words_to_cut // 2

    while words_to_cut_before > len(sequences[0]):
        words_to_cut_before -= len(sequences[0])
        sequences = sequences[1:]
    sequences[0] = sequences[0][words_to_cut_before:]

    while words_to_cut_after > len(sequences[-1]):
        words_to_cut_after -= len(sequences[-1])
        sequences = sequences[:-1]
    last = len(sequences[-1]) - words_to_cut_after
    sequences[-1] = sequences[-1][:last]

    return sequences


def pad_ids(arrays, padding, max_length=-1):
    if max_length < 0:
        max_length = max(list(map(len, arrays)))

    arrays = [array + [padding] * (max_length - len(array)) for array in arrays]

    return arrays


def text_to_list(text):
    return [t.strip().strip("'") for t in text.strip("]").strip("[").split(",")]



import torch
from math import log
from itertools import chain
from collections import defaultdict, Counter
from multiprocessing import Pool
from functools import partial
from tqdm.auto import tqdm

__all__ = ["bert_types"]

bert_types = [
    "bert-base-uncased",
    "bert-large-uncased",
    "bert-base-cased",
    "bert-large-cased",
    "bert-base-multilingual-uncased",
    "bert-base-multilingual-cased",
    "bert-base-chinese",
]


def padding(arr, pad_token, dtype=torch.long):
    lens = torch.LongTensor([len(a) for a in arr])
    max_len = lens.max().item()
    padded = torch.ones(len(arr), max_len, dtype=dtype) * pad_token
    mask = torch.zeros(len(arr), max_len, dtype=torch.long)
    for i, a in enumerate(arr):
        padded[i, : lens[i]] = torch.tensor(a, dtype=dtype)
        mask[i, : lens[i]] = 1
    return padded, lens, mask


def bert_encode(model, x, attention_mask):
    model.eval()
    with torch.no_grad():
        x_encoded_layers, _ = model(x, attention_mask=attention_mask, return_dict=False)
    return x_encoded_layers


def process(a, tokenizer=None):
    if tokenizer is not None:
        a = ["[CLS]"] + tokenizer.tokenize(a) + ["[SEP]"]
        a = tokenizer.convert_tokens_to_ids(a)
    return set(a)


def get_idf_dict(arr, tokenizer, nthreads=4):
    """
    Returns mapping from word piece index to its inverse document frequency.
    Args:
        - :param: `arr` (list of str) : sentences to process.
        - :param: `tokenizer` : a BERT tokenizer corresponds to `model`.
        - :param: `nthreads` (int) : number of CPU threads to use
    """
    idf_count = Counter()
    num_docs = len(arr)

    process_partial = partial(process, tokenizer=tokenizer)

    with Pool(nthreads) as p:
        idf_count.update(chain.from_iterable(p.map(process_partial, arr)))

    idf_dict = defaultdict(lambda: log((num_docs + 1) / (1)))
    idf_dict.update(
        {idx: log((num_docs + 1) / (c + 1)) for (idx, c) in idf_count.items()}
    )
    return idf_dict


def collate_idf(arr, tokenize, numericalize, idf_dict, pad="[PAD]", device="cuda:0"):
    """
    Helper function that pads a list of sentences to hvae the same length and
    loads idf score for words in the sentences.
    Args:
        - :param: `arr` (list of str): sentences to process.
        - :param: `tokenize` : a function that takes a string and return list
                  of tokens.
        - :param: `numericalize` : a function that takes a list of tokens and
                  return list of token indexes.
        - :param: `idf_dict` (dict): mapping a word piece index to its
                               inverse document frequency
        - :param: `pad` (str): the padding token.
        - :param: `device` (str): device to use, e.g. 'cpu' or 'cuda'
    """
    arr = [["[CLS]"] + tokenize(a) + ["[SEP]"] for a in arr]
    arr = [numericalize(a) for a in arr]

    idf_weights = [[idf_dict[i] for i in a] for a in arr]

    pad_token = numericalize([pad])[0]

    padded, lens, mask = padding(arr, pad_token, dtype=torch.long)
    padded_idf, _, _ = padding(idf_weights, pad_token, dtype=torch.float)

    padded = padded.to(device=device)
    mask = mask.to(device=device)
    lens = lens.to(device=device)
    return padded, padded_idf, lens, mask


def get_bert_embedding(
    all_sens, model, tokenizer, idf_dict, batch_size=-1, device="cuda:0"
):
    """
    Compute BERT embedding in batches.
    Args:
        - :param: `all_sens` (list of str) : sentences to encode.
        - :param: `model` : a BERT model from `pytorch_pretrained_bert`.
        - :param: `tokenizer` : a BERT tokenizer corresponds to `model`.
        - :param: `idf_dict` (dict) : mapping a word piece index to its
                               inverse document frequency
        - :param: `device` (str): device to use, e.g. 'cpu' or 'cuda'
    """

    padded_sens, padded_idf, lens, mask = collate_idf(
        all_sens,
        tokenizer.tokenize,
        tokenizer.convert_tokens_to_ids,
        idf_dict,
        device=device,
    )

    if batch_size == -1:
        batch_size = len(all_sens)

    embeddings = []
    with torch.no_grad():
        for i in range(0, len(all_sens), batch_size):
            batch_embedding = bert_encode(
                model,
                padded_sens[i : i + batch_size],
                attention_mask=mask[i : i + batch_size],
            )
            # batch_embedding = torch.stack(batch_embedding)
            embeddings.append(batch_embedding)
            del batch_embedding

    total_embedding = torch.cat(embeddings, dim=0)

    return total_embedding, lens, mask, padded_idf


def greedy_cos_idf(
    ref_embedding,
    ref_lens,
    ref_masks,
    ref_idf,
    hyp_embedding,
    hyp_lens,
    hyp_masks,
    hyp_idf,
):
    """
    Compute greedy matching based on cosine similarity.
    Args:
        - :param: `ref_embedding` (torch.Tensor):
                   embeddings of reference sentences, BxKxd,
                   B: batch size, K: longest length, d: bert dimenison
        - :param: `ref_lens` (list of int): list of reference sentence length.
        - :param: `ref_masks` (torch.LongTensor): BxKxK, BERT attention mask for
                   reference sentences.
        - :param: `ref_idf` (torch.Tensor): BxK, idf score of each word
                   piece in the reference setence
        - :param: `hyp_embedding` (torch.Tensor):
                   embeddings of candidate sentences, BxKxd,
                   B: batch size, K: longest length, d: bert dimenison
        - :param: `hyp_lens` (list of int): list of candidate sentence length.
        - :param: `hyp_masks` (torch.LongTensor): BxKxK, BERT attention mask for
                   candidate sentences.
        - :param: `hyp_idf` (torch.Tensor): BxK, idf score of each word
                   piece in the candidate setence
    """

    ref_embedding.div_(torch.norm(ref_embedding, dim=-1).unsqueeze(-1))
    hyp_embedding.div_(torch.norm(hyp_embedding, dim=-1).unsqueeze(-1))

    batch_size = ref_embedding.size(0)

    sim = torch.bmm(hyp_embedding, ref_embedding.transpose(1, 2))
    masks = torch.bmm(hyp_masks.unsqueeze(2).float(), ref_masks.unsqueeze(1).float())
    masks = (
        masks.expand(batch_size, masks.size(1), masks.size(2)).contiguous().view_as(sim)
    )

    masks = masks.float().to(sim.device)
    sim = sim * masks

    word_precision = sim.max(dim=2)[0]
    word_recall = sim.max(dim=1)[0]

    hyp_idf.div_(hyp_idf.sum(dim=1, keepdim=True))
    ref_idf.div_(ref_idf.sum(dim=1, keepdim=True))
    precision_scale = hyp_idf.to(word_precision.device)
    recall_scale = ref_idf.to(word_recall.device)
    P = (word_precision * precision_scale).sum(dim=1)
    R = (word_recall * recall_scale).sum(dim=1)

    F = 2 * P * R / (P + R)
    return P, R, F


def bert_cos_score_idf(
    model,
    refs,
    hyps,
    tokenizer,
    idf_dict,
    verbose=False,
    batch_size=64,
    device="cuda:0",
):
    """
    Compute BERTScore.
    Args:
        - :param: `model` : a BERT model in `pytorch_pretrained_bert`
        - :param: `refs` (list of str): reference sentences
        - :param: `hyps` (list of str): candidate sentences
        - :param: `tokenzier` : a BERT tokenizer corresponds to `model`
        - :param: `idf_dict` : a dictionary mapping a word piece index to its
                               inverse document frequency
        - :param: `verbose` (bool): turn on intermediate status update
        - :param: `batch_size` (int): bert score processing batch size
        - :param: `device` (str): device to use, e.g. 'cpu' or 'cuda'
    """
    preds = []
    iter_range = range(0, len(refs), batch_size)
    if verbose:
        iter_range = tqdm(iter_range)
    for batch_start in iter_range:
        batch_refs = refs[batch_start : batch_start + batch_size]
        batch_hyps = hyps[batch_start : batch_start + batch_size]
        ref_stats = get_bert_embedding(
            batch_refs, model, tokenizer, idf_dict, device=device
        )
        hyp_stats = get_bert_embedding(
            batch_hyps, model, tokenizer, idf_dict, device=device
        )

        P, R, F1 = greedy_cos_idf(*ref_stats, *hyp_stats)
        preds.append(torch.stack((P, R, F1), dim=1).cpu())
    preds = torch.cat(preds, dim=0)
    return preds



import numpy as np
import torch
from torch import cuda
from transformers import GPT2Tokenizer, GPT2LMHeadModel

from kogito.core.model import KnowledgeModel
from kogito.core.knowledge import KnowledgeGraph

device = "cuda" if cuda.is_available() else "cpu"


class GPT2Zeroshot(KnowledgeModel):
    """Zeroshot knowledge model based on GPT-2"""

    def __init__(self, gpt2_model: str = "gpt2") -> None:
        """Initialize GPT-2 model
        Args:
            gpt2_model (str, optional): HuggingFace model name for gpt2. Defaults to "gpt2".
        """
        self.tokenizer = GPT2Tokenizer.from_pretrained(gpt2_model)
        self.model = GPT2LMHeadModel.from_pretrained(gpt2_model)
        self.model.to(device)

    def train(self):
        raise ValueError("GPT-2 Zeroshot model is not trainable")

    def save_pretrained(self, save_path):
        self.model.save_pretrained(save_path)
        self.tokenizer.save_pretrained(save_path)

    @classmethod
    def from_pretrained(cls, model_name_or_path: str = "gpt2"):
        return cls(model_name_or_path)

    def generate(
        self, input_graph: KnowledgeGraph, seed: int = 42, **kwargs
    ) -> KnowledgeGraph:
        """Generate inferences from GPT2 model
        Args:
            input_graph (KnowledgeGraph): Input dataset
            seed (int, optional): Random seed. Defaults to 42.
            kwargs: Additional arguments to pass to the model.generate() function
        Returns:
            KnowledgeGraph: Completed knowledge graph
        """
        torch.manual_seed(seed)
        np.random.seed(seed)
        torch.backends.cudnn.deterministic = True

        if "top_k" not in kwargs:
            kwargs["top_k"] = 1

        if "top_p" not in kwargs:
            kwargs["top_p"] = 0.9

        if "num_return_sequences" not in kwargs:
            kwargs["num_return_sequences"] = 3

        if "num_beams" not in kwargs:
            kwargs["num_beams"] = 3

        if "temperature" not in kwargs:
            kwargs["temperature"] = 0.7

        if "repetition_penalty" not in kwargs:
            kwargs["repetition_penalty"] = 1.2

        if "max_length" not in kwargs:
            kwargs["max_length"] = 32

        if "do_sample" not in kwargs:
            kwargs["do_sample"] = True

        outputs = []
        for input_kg in input_graph:
            prompt = input_kg.to_prompt()
            input_ids = self.tokenizer.encode(
                prompt, add_special_tokens=False, return_tensors="pt"
            )
            input_length = input_ids.size(1)
            generations = self.model.generate(
                input_ids=input_ids.to(device),
                max_length=input_length + kwargs["max_length"],
                eos_token_id=198,
                **kwargs
            )

            if len(generations.shape) > 2:
                generations.squeeze_()

            text_generations = []
            for gen in generations:
                gen = gen.tolist()
                text = self.tokenizer.decode(
                    gen[input_length:], clean_up_tokenization_spaces=True
                )
                text_generations.append(text.strip())

            output_kg = input_kg.copy()
            output_kg.tails = text_generations
            outputs.append(output_kg)

        return KnowledgeGraph(outputs)



from typing import Union, List, Optional
import warnings

import spacy

from kogito.core.knowledge import Knowledge, KnowledgeGraph
from kogito.core.head import KnowledgeHead
from kogito.core.processors.head import (
    KnowledgeHeadExtractor,
    SentenceHeadExtractor,
    NounPhraseHeadExtractor,
    VerbPhraseHeadExtractor,
)
from kogito.core.relation import KnowledgeRelation
from kogito.core.processors.relation import (
    GraphBasedRelationMatcher,
    KnowledgeRelationMatcher,
    SimpleRelationMatcher,
    BaseRelationMatcher,
)
from kogito.core.model import KnowledgeModel
from kogito.core.linker import KnowledgeLinker
from kogito.models.gpt3.zeroshot import GPT3Zeroshot
from kogito.linkers.deberta import DebertaLinker


class CommonsenseInference:
    """Main interface for commonsense inference"""

    def __init__(self, language: str = "en_core_web_sm") -> None:
        """Initialize a commonsense inference module

        Args:
            language (str, optional): Spacy language pipeline to use. Defaults to "en_core_web_sm".
        """
        self.language = language
        self.nlp = spacy.load(language, exclude=["ner"])

        self._head_processors = {
            "sentence_extractor": SentenceHeadExtractor("sentence_extractor", self.nlp),
            "noun_phrase_extractor": NounPhraseHeadExtractor(
                "noun_phrase_extractor", self.nlp
            ),
            "verb_phrase_extractor": VerbPhraseHeadExtractor(
                "verb_phrase_extractor", self.nlp
            ),
        }
        self._relation_processors = {
            "simple_relation_matcher": SimpleRelationMatcher(
                "simple_matcher", self.nlp
            ),
            "graph_relation_matcher": GraphBasedRelationMatcher(
                "graph_matcher", self.nlp
            ),
        }

    @property
    def processors(self) -> dict:
        """List all processors

        Returns:
            dict: List of head and relation processors
        """
        return {
            "head": list(self._head_processors.keys()),
            "relation": list(self._relation_processors.keys()),
        }

    def infer(
        self,
        text: Optional[str] = None,
        model: Optional[KnowledgeModel] = None,
        heads: Optional[List[str]] = None,
        model_args: Optional[dict] = None,
        extract_heads: bool = True,
        match_relations: bool = True,
        relations: Optional[List[KnowledgeRelation]] = None,
        dry_run: bool = False,
        sample_graph: Optional[KnowledgeGraph] = None,
        context: Optional[Union[List[str], str]] = None,
        linker: Optional[KnowledgeLinker] = None,
        threshold: float = 0.5,
    ) -> KnowledgeGraph:
        """Make commonsense inferences.

        Args:
            text (Optional[str], optional): Text to use to extract commonsense inferences from.
                If omitted, no head extraction will be performed even if ``extract_heads`` is True.
                If provided and ``extract_heads`` is False, text will be used as a head as is.
                Defaults to None.
            model (Optional[KnowledgeModel], optional): Knowledge model to use for inference.
                If omitted, behaviour is equivalent to dry-run mode, i.e. no inference will be performed
                and incomplete input graph will be returned.
                Defaults to None.
            heads (Optional[List[str]], optional): List of custom heads to use for inference. Defaults to None.
            model_args (Optional[dict], optional): Custom arguments to pass to ``KnowledgeModel.generate()`` method.
                Defaults to None.
            extract_heads (bool, optional): Whether to extract heads from given text if any. Defaults to True.
            match_relations (bool, optional): Whether to do smart relation matching. Defaults to True.
            relations (Optional[List[KnowledgeRelation]], optional): Subset of relations to use for direct matching.
                If ``match_relations`` is true, intersection of matched and given relations will be used.
                Defaults to None.
            dry_run (bool, optional): Whether to skip actual inference and return incomplete input graph.
                Defaults to False.
            sample_graph (Optional[KnowledgeGraph], optional): A knowledge graph containing examples.
                It can be used to provide examples for GPT-3 inference. If omitted and GPT-3 model is used,
                warning will be raised.
                Defaults to None.
            context (Optional[Union[List[str], str]], optional): Context text. Can be either given as a list of
                                            sentences or as a string, in which case, it will be
                                            split into sentences using spacy engine. Defaults to None.
            threshold (float, optional): Relevance probability used for filtering. Defaults to 0.5.
            linker (Optional[KnowledgeLinker], optional): Knowledge linker model used for linking to given context.
                                                            Defaults to Deberta-based linker.

        Raises:
            ValueError: if relations argument is not of type list

        Returns:
            KnowledgeGraph: Inferred knowledge graph.
        """
        kg_heads = []
        head_relations = set()
        head_texts = set()
        model_args = model_args or {}

        if relations is not None and not isinstance(relations, list):
            raise ValueError("Relation subset should be a list")

        if text is not None:
            if not isinstance(text, str):
                raise ValueError("Text should be a string")
            text = text.strip()

        if heads is not None:
            if not isinstance(heads, list):
                raise ValueError("Heads should be a list")

        if not text and not heads:
            warnings.warn("Skipping inference, no text or head provided")
            return None

        if heads:
            for head in heads:
                if isinstance(head, str) and head.strip():
                    head_texts.add(head)
                    kg_heads.append(KnowledgeHead(text=head))

        if extract_heads:
            if text.strip():
                print("Extracting knowledge heads...")
                for head_proc in self._head_processors.values():
                    extracted_heads = head_proc.extract(text)
                    for head in extracted_heads:
                        head_text = head.text.strip().lower()
                        # Check for duplication
                        if head_text not in head_texts:
                            kg_heads.append(head)
                            head_texts.add(head_text)
        else:
            if text and text.strip() not in head_texts:
                head_texts.add(text)
                kg_heads.append(KnowledgeHead(text=text))

        if not kg_heads:
            warnings.warn("Skipping inference, no heads found.")
            return None

        base_relation_matcher = BaseRelationMatcher("base-relation-matcher")

        if match_relations:
            print("Matching knowledge heads with relations...")
            for relation_proc in self._relation_processors.values():
                head_relations = head_relations.union(
                    set(
                        relation_proc.match(
                            kg_heads, relations, sample_graph=sample_graph
                        )
                    )
                )
        else:
            print("Pairing knowledge heads with all relations...")
            head_relations = head_relations.union(
                set(
                    base_relation_matcher.match(
                        kg_heads, relations, sample_graph=sample_graph
                    )
                )
            )

        kg_list = []

        for head_relation in head_relations:
            head, relation = head_relation
            kg_list.append(Knowledge(head=head, relation=relation))

        input_graph = KnowledgeGraph(kg_list)

        if sample_graph:
            input_graph = input_graph + sample_graph
        else:
            if isinstance(model, GPT3Zeroshot):
                warnings.warn(
                    "Sample graph not found, but recommended for good performance with GPT-3 based inference."
                )

        if dry_run or not model:
            return input_graph.sort()

        print("Generating knowledge graph...")
        output_graph = model.generate(input_graph, **model_args)

        if context:
            if not linker:
                print("Loading default Deberta linker for filtering...")
                linker = DebertaLinker()

            print("Filtering knowledge graph based on the context...")
            output_graph = linker.filter(output_graph, context, threshold=threshold)

        output_graph.clean()
        output_graph.sort()

        return output_graph

    def add_processor(
        self, processor: Union[KnowledgeHeadExtractor, KnowledgeRelationMatcher]
    ) -> None:
        """Add a new head or relation processor to the module

        Args:
            processor (Union[KnowledgeHeadExtractor, KnowledgeRelationMatcher]): Head or relation processor.

        Raises:
            ValueError: When processor type is not recognized.
        """
        if isinstance(processor, KnowledgeHeadExtractor):
            self._head_processors[processor.name] = processor
            processor.lang = self.nlp
        elif isinstance(processor, KnowledgeRelationMatcher):
            self._relation_processors[processor.name] = processor
            processor.lang = self.nlp
        else:
            raise ValueError("Unknown processor")

    def remove_processor(self, processor_name: str) -> None:
        """Remove a processor from the module

        Args:
            processor_name (str): Name of the processor to remove
        """
        if processor_name in self._head_processors:
            del self._head_processors[processor_name]
        elif processor_name in self._relation_processors:
            del self._relation_processors[processor_name]



from typing import List, Union, Optional
import pandas as pd
import json

from kogito.core.relation import KnowledgeRelation, KnowledgeRelationType, RELATION_SIZE
from kogito.core.head import KnowledgeHead
from kogito.core.utils import text_to_list

EOS_TOKEN = "[EOS]"
GEN_TOKEN = "[GEN]"
PAD_TOKEN = "[PAD]"

DECODE_METHODS = ["greedy", "beam"]


class Knowledge:
    """
    Represents a concept of Knowledge
    """

    def __init__(
        self,
        head: Optional[Union[KnowledgeHead, str]] = None,
        relation: Optional[Union[KnowledgeRelation, str]] = None,
        tails: Optional[List[str]] = None,
    ) -> None:
        """Initialize a Knowledge instance.

        Args:
            head (Optional[Union[KnowledgeHead, str]], optional): Instance of a knowledge head. Defaults to None.
            relation (Optional[Union[KnowledgeRelation, str]], optional): Instance of a knowledge relation.
                                                                         Defaults to None.
            tails (Optional[List[str]], optional): List of knowledge tails. Defaults to None.
        """
        self.head = head if isinstance(head, KnowledgeHead) else KnowledgeHead(head)
        self.relation = (
            relation
            if isinstance(relation, KnowledgeRelation)
            else KnowledgeRelation.from_text(relation)
        )
        self.tails = tails or []
        if not isinstance(self.tails, (list, tuple)):
            tails = str(self.tails).strip()
            if tails.startswith("[") and tails.endswith("]"):
                tails = text_to_list(tails)
                self.tails = tails
            else:
                self.tails = [tails]
        else:
            self.tails = list(self.tails)

    def __repr__(self) -> str:
        return f'Knowledge(head="{str(self.head)}", relation="{str(self.relation)}", tails={self.tails})'

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Knowledge)
            and self.head == other.head
            and self.relation == other.relation
            and set(self.tails) == set(other.tails)
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self.head, self.relation, tuple(self.tails)))

    def to_prompt(self, include_tail: bool = False, **kwargs) -> str:
        """Convert knowledge to a prompt text.

        Args:
            include_tail (bool, optional): Whether to include tails in the prompt. Defaults to False.

        Returns:
            str: Prompt text for the knowledge
        """
        head = self.head
        relation = self.relation
        tail = self.tails[0] if self.tails else None
        return relation.verbalize(str(head), tail, include_tail=include_tail, **kwargs)

    def copy(self) -> "Knowledge":
        """Copy itself.

        Returns:
            Knowledge: Copied knowledge
        """
        return Knowledge(
            head=self.head.copy(),
            relation=self.relation.copy(),
            tails=self.tails.copy(),
        )

    def to_json(self, only_one_tail: bool = False) -> dict:
        """Convert knowledge to dictionary

        Args:
            only_one_tail (bool, optional): Include only one tail. Defaults to False.

        Returns:
            dict: Jsonified knowledge
        """
        tails = self.tails

        if only_one_tail:
            if self.tails:
                tails = self.tails[0]
            else:
                tails = ""

        return {
            "head": str(self.head),
            "relation": str(self.relation),
            "tails": tails,
        }


class KnowledgeGraph:
    """
    Represents a concept of Knowledge Graph.
    """

    def __init__(self, graph: List[Knowledge]) -> None:
        """Initialize a knowledge graph

        Args:
            graph (List[Knowledge]): List of Knowledge instances
        """
        self.graph = graph
        self._graph_iter = None

    def __iter__(self):
        self._graph_iter = iter(self.graph)
        return self

    def __next__(self):
        return next(self._graph_iter)

    def __len__(self):
        return len(self.graph)

    def __getitem__(self, idx):
        return self.graph[idx]

    def __add__(self, other):
        return self.union(other)

    def __or__(self, other):
        return self.union(other)

    def __and__(self, other):
        return self.intersection(other)

    def __sub__(self, other):
        return self.difference(other)

    def __repr__(self) -> str:
        return "\n".join([str(kg) for kg in self.graph])

    @classmethod
    def from_jsonl(
        cls,
        filepath: str,
        head_attr: str = "head",
        relation_attr: str = "relation",
        tails_attr: str = "tails",
        relation_type: KnowledgeRelationType = KnowledgeRelationType.ATOMIC,
    ) -> "KnowledgeGraph":
        """Initialize a knowledge graph from json file.

        Args:
            filepath (str): Path to the graph file.
            head_attr (str, optional): JSON attribute for head. Defaults to "head".
            relation_attr (str, optional): JSON attribute for relation. Defaults to "relation".
            tails_attr (str, optional): JSON attribute for tails. Defaults to "tails".
            relation_type (KnowledgeRelationType, optional): Type of relation to use.
                                                            Defaults to KnowledgeRelationType.ATOMIC.

        Returns:
            KnowledgeGraph: An instance of KnowledgeGraph
        """
        kg_list = []

        with open(filepath) as file:
            for line in file:
                kg_json = json.loads(line)
                head = kg_json.get(head_attr)
                relation = KnowledgeRelation.from_text(
                    kg_json.get(relation_attr), relation_type
                )
                tails = kg_json.get(tails_attr)
                kg_list.append(Knowledge(head=head, relation=relation, tails=tails))

        return cls(kg_list)

    @classmethod
    def from_csv(
        cls,
        filepath: str,
        header: Optional[Union[int, List[int], str]] = "infer",
        head_col: str = "head",
        relation_col: str = "relation",
        tails_col: str = "tails",
        sep: str = ",",
        relation_type: KnowledgeRelationType = KnowledgeRelationType.ATOMIC,
    ) -> "KnowledgeGraph":
        """Initialize a knowledge graph from csv file.

        Args:
            filepath (str): Path to the graph file.
            header (Union[int, List[int], str], optional): Whether to look for header. Defaults to "infer".
            head_col (str, optional): Head column name. Defaults to "head".
            relation_col (str, optional): Relation column name. Defaults to "relation".
            tails_col (str, optional): Tails column name. Defaults to "tails".
            sep (str, optional): Delimiter to use. Defaults to ",".
            relation_type (KnowledgeRelationType, optional): Relation type to use.
                                                            Defaults to KnowledgeRelationType.ATOMIC.

        Returns:
            KnowledgeGraph: An instance of KnowledgeGraph
        """
        graph_df = pd.read_csv(
            filepath,
            sep=sep,
            header=header,
            usecols=[head_col, relation_col, tails_col],
        )
        return cls.from_dataframe(
            df=graph_df,
            head_col=head_col,
            relation_col=relation_col,
            tails_col=tails_col,
            relation_type=relation_type,
        )

    @classmethod
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        head_col: str = "head",
        relation_col: str = "relation",
        tails_col: str = "tails",
        relation_type: KnowledgeRelationType = KnowledgeRelationType.ATOMIC,
    ) -> "KnowledgeGraph":
        """Initialize a knowledge graph from csv file.

        Args:
            df (pd.DataFrame): Graph dataframe.
            head_col (str, optional): Head column name. Defaults to "head".
            relation_col (str, optional): Relation column name. Defaults to "relation".
            tails_col (str, optional): Tails column name. Defaults to "tails".
            relation_type (KnowledgeRelationType, optional): Relation type to use.
                                                            Defaults to KnowledgeRelationType.ATOMIC.

        Returns:
            KnowledgeGraph: An instance of KnowledgeGraph
        """
        kg_list = []
        exists_head_col = head_col in df.columns
        exists_relation_col = relation_col in df.columns
        exists_tails_col = tails_col in df.columns

        for _, row in df.iterrows():
            head = row[head_col].strip() if exists_head_col else None
            relation = (
                KnowledgeRelation.from_text(row[relation_col].strip(), relation_type)
                if exists_relation_col
                else None
            )
            tails = row[tails_col] if exists_tails_col else None
            kg_list.append(Knowledge(head=head, relation=relation, tails=tails))

        return cls(kg_list)

    def to_jsonl(self, filepath: str) -> None:
        """Write knowledge graph to a json file

        Args:
            filepath (str): JSON file path
        """
        with open(filepath, "w") as file:
            lines = []
            for kg in self.graph:
                lines.append(json.dumps(kg.to_json()))
            file.writelines("\n".join(lines))

    def to_dataframe(self) -> pd.DataFrame:
        """Convert knowledge graph to a pandas dataframe

        Returns:
            pd.DataFrame: Pandas dataframe of a knowledge graph
        """
        return pd.DataFrame([kg.to_json(only_one_tail=True) for kg in self.graph])

    def union(self, other: "KnowledgeGraph") -> "KnowledgeGraph":
        """Union two knowledge graphs

        Args:
            other (KnowledgeGraph): Knowledge graph to union with.

        Returns:
            KnowledgeGraph: Merged knowledge graph
        """
        return KnowledgeGraph(set(self.graph).union(set(other.graph)))

    def intersection(self, other: "KnowledgeGraph") -> "KnowledgeGraph":
        """Intersect two knowledge graphs

        Args:
            other (KnowledgeGraph): Knowledge graph to intersect with.

        Returns:
            KnowledgeGraph: Intersection of two graphs
        """
        return KnowledgeGraph(set(self.graph).intersection(set(other.graph)))

    def difference(self, other: "KnowledgeGraph") -> "KnowledgeGraph":
        """Subtract knowledge graphs

        Args:
            other (KnowledgeGraph): Knowledge graph to subtract.

        Returns:
            KnowledgeGraph: Difference of two graphs
        """
        return KnowledgeGraph(set(self.graph).difference(set(other.graph)))

    def sort(self) -> "KnowledgeGraph":
        """Sort graph based on head text and
        relation distribution in ATOMIC.

        Returns:
            KnowledgeGraph: Sorted graph.
        """
        self.graph.sort(
            key=lambda kg: (kg.head.text, RELATION_SIZE.get(kg.relation.text, 0)),
            reverse=True,
        )
        return self

    def clean(self) -> "KnowledgeGraph":
        """Clean graph tails by removing empty generations.

        Returns:
            KnowledgeGraph: Cleaned graph.
        """
        clean_graph = []

        for kg in self.graph:
            if kg.tails:
                clean_tails = []
                for tail in kg.tails:
                    clean_tail = tail.strip()
                    if clean_tail:
                        clean_tails.append(clean_tail)
                if clean_tails:
                    kg.tails = clean_tails
                    clean_graph.append(kg)

        self.graph = clean_graph
        return self



from typing import Any, Callable, Optional
from enum import Enum


class KnowledgeHeadType(Enum):
    """
    Type of a Knowledge Head
    """

    TEXT = "text"
    SENTENCE = "sentence"
    NOUN_PHRASE = "noun_phrase"
    VERB_PHRASE = "verb_phrase"


class KnowledgeHead:
    """
    Represents a concept of Knowledge Head.
    """

    def __init__(
        self,
        text: str,
        type: KnowledgeHeadType = KnowledgeHeadType.TEXT,
        entity: Any = None,
        verbalizer: Optional[Callable] = None,
    ) -> None:
        """Initialize a Knowledge Head.

        Args:
            text (str): Head text.
            type (KnowledgeHeadType, optional): Type of a Knowledge head. Defaults to KnowledgeHeadType.TEXT.
            entity (Any, optional): External Knowledge head entity. Defaults to None.
            verbalizer (Optional[Callable], optional): Function to convert knowledge head to natural text.
                                                      Defaults to None.
        """
        self.text = text.strip()
        self.type = type
        self.entity = entity
        self.verbalizer = verbalizer

    def __eq__(self, other: object) -> bool:
        return isinstance(other, KnowledgeHead) and self.text == other.text

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.text)

    def verbalize(self) -> Optional[str]:
        """Convert head to a meaningful text.

        Returns:
            Optional[str]: Verbalized head
        """
        if self.verbalizer:
            return self.verbalizer(self.text)

    def __repr__(self) -> str:
        return str(self.text)

    def copy(self) -> "KnowledgeHead":
        """Copy itself

        Returns:
            KnowledgeHead: Copied knowledge head
        """
        return KnowledgeHead(
            text=self.text,
            type=self.type,
            entity=self.entity,
            verbalizer=self.verbalizer,
        )


def head_verbalizer(head: str):
    return head



#!/usr/bin/env python
#
# File Name : rouge.py
#
# Description : Computes ROUGE-L metric as described by Lin and Hovey (2004)
#
# Creation Date : 2015-01-07 06:03
# Author : Ramakrishna Vedantam <vrama91@vt.edu>

import numpy as np


def my_lcs(string, sub):
    """
    Calculates longest common subsequence for a pair of tokenized strings
    :param string : list of str : tokens from a string split using whitespace
    :param sub : list of str : shorter string, also split using whitespace
    :returns: length (list of int): length of the longest common subsequence between the two strings

    Note: my_lcs only gives length of the longest common subsequence, not the actual LCS
    """
    if len(string) < len(sub):
        sub, string = string, sub

    lengths = [[0 for i in range(0, len(sub) + 1)] for j in range(0, len(string) + 1)]

    for j in range(1, len(sub) + 1):
        for i in range(1, len(string) + 1):
            if string[i - 1] == sub[j - 1]:
                lengths[i][j] = lengths[i - 1][j - 1] + 1
            else:
                lengths[i][j] = max(lengths[i - 1][j], lengths[i][j - 1])

    return lengths[len(string)][len(sub)]


class Rouge:
    """
    Class for computing ROUGE-L score for a set of candidate sentences for the MS COCO test set

    """

    def __init__(self):
        # vrama91: updated the value below based on discussion with Hovey
        self.beta = 1.2

    def calc_score(self, candidate, refs):
        """
        Compute ROUGE-L score given one candidate and references for an image
        :param candidate: str : candidate sentence to be evaluated
        :param refs: list of str : COCO reference sentences for the particular image to be evaluated
        :returns score: int (ROUGE-L score for the candidate evaluated against references)
        """
        assert len(candidate) == 1
        assert len(refs) > 0
        prec = []
        rec = []

        # split into tokens
        token_c = candidate[0].split(" ")

        for reference in refs:
            # split into tokens
            token_r = reference.split(" ")
            # compute the longest common subsequence
            lcs = my_lcs(token_r, token_c)
            prec.append(lcs / float(len(token_c)))
            rec.append(lcs / float(len(token_r)))

        prec_max = max(prec)
        rec_max = max(rec)

        if prec_max != 0 and rec_max != 0:
            score = ((1 + self.beta**2) * prec_max * rec_max) / float(
                rec_max + self.beta**2 * prec_max
            )
        else:
            score = 0.0
        return score

    def compute_score(self, gts, res):
        """
        Computes Rouge-L score given a set of reference and candidate sentences for the dataset
        Invoked by evaluate_captions.py
        :param hypo_for_image: dict : candidate / test sentences with "image name" key and
                    "tokenized sentences" as values
        :param ref_for_image: dict : reference MS-COCO sentences with "image name" key and
                    "tokenized sentences" as values
        :returns: average_score: float (mean ROUGE-L score computed by averaging scores for all the images)
        """
        assert gts.keys() == res.keys()
        imgIds = gts.keys()

        score = []
        for id in imgIds:
            hypo = res[id]
            ref = gts[id]

            score.append(self.calc_score(hypo, ref))

            # Sanity check.
            assert type(hypo) is list
            assert len(hypo) == 1
            assert type(ref) is list
            assert len(ref) > 0

        average_score = np.mean(np.array(score))
        return average_score, np.array(score)

    def method(self):
        return "Rouge"



from typing import List, Union, Tuple
from abc import ABC, abstractmethod, abstractclassmethod

from kogito.core.knowledge import KnowledgeGraph


class KnowledgeLinker(ABC):
    """Base Knowledge Linker"""

    @abstractmethod
    def save_pretrained(self, save_path: str) -> None:
        """Save linker as a pretrained model

        Args:
            save_path (str): Directory to save the linker to.

        Raises:
            NotImplementedError: This method has to be implemented by
                                 concrete subclasses.
        """
        raise NotImplementedError

    @abstractclassmethod
    def from_pretrained(cls, model_name_or_path: str) -> "KnowledgeLinker":
        """Load model from a pretrained model path
        This method can load linkers either from HuggingFace by model name
        or from disk by model path.

        Args:
            model_name_or_path (str): HuggingFace model name or local model path to load from.

        Raises:
            NotImplementedError: This method has to be implemented by
                                 concrete subclasses.

        Returns:
            KnowledgeLinker: Loaded knowledge linker.
        """
        raise NotImplementedError

    @abstractmethod
    def link(
        self, input_graph: KnowledgeGraph, context: Union[List[str], str]
    ) -> List[List[float]]:
        """Link given knowledge graph with the context.
        This method computes a relevance probability for each knowledge in the graph
        with respect to the given context and returns these probabilities in a list
        in the same order as the knowledge tuples are in the given graph. Note that returned object
        is a list of list of numbers because a knowledge tuple might have multiple tails and the probability
        is calculated for each combination.

        Args:
            input_graph (KnowledgeGraph): Input graph to link.
            context (Union[List[str], str]): Context text. Can be either given as a list of
                                            sentences or as a string, in which case, it will be
                                            split into sentences using spacy engine.

        Returns:
            List[List[float]]: List of relevance probabilities for each tail
        """
        raise NotImplementedError

    def filter(
        self,
        input_graph: KnowledgeGraph,
        context: Union[List[str], str],
        threshold: float = 0.5,
        return_probs: bool = False,
    ) -> Union[KnowledgeGraph, Tuple[KnowledgeGraph, List[List[float]]]]:
        """Filter given graph based on context relevancy.
        This method under the hood links the graph to the context and then filters knowledge tuples from the graph
        which have a relevance probability lower than the given threshold. Filtered knowledge tuples
        are returned as a new knowledge graph. If there are multiple tails for a given knowledge, these tails will be
        filtered as well.

        Args:
            input_graph (KnowledgeGraph): Input graph to filter.
            context (Union[List[str], str]): Context text. Can be either given as a list of
                                            sentences or as a string, in which case, it will be
                                            split into sentences using spacy engine.
            threshold (float, optional): Relevance probability used for filtering. Defaults to 0.5.
            return_probs (bool, optional): Whether to return all the relevancy probs for the input graph.
                                            Defaults to False.
        Returns:
            Union[KnowledgeGraph, Tuple[KnowledgeGraph, List[List[float]]]]:
            Filtered knowledge graph based on the relevancy scores and optionally, the relevancy scores.
        """
        probs = self.link(input_graph, context)
        filtered_kgs = []

        for kg, tail_probs in zip(input_graph, probs):
            filtered_tails = []

            for i, prob in enumerate(tail_probs):
                if prob >= threshold:
                    filtered_tails.append(kg.tails[i])

            if filtered_tails:
                kg.tails = filtered_tails
                filtered_kgs.append(kg)

        output_graph = KnowledgeGraph(filtered_kgs)

        if return_probs:
            return output_graph, probs

        return output_graph



import torchmetrics
import numpy as np
import spacy


def text_to_embedding(text, vocab, embedding_matrix, pooling="max", lang=None):
    if not lang:
        lang = spacy.load("en_core_web_sm")

    doc = lang(text)
    vectors = []
    for token in doc:
        if token.text in vocab:
            vectors.append(embedding_matrix[vocab[token.text]])

    if vectors:
        if pooling == "max":
            return np.amax(np.array(vectors, dtype=np.float32), axis=0)
        return np.mean(vectors, axis=0, dtype=np.float32)


class Evaluator:
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.metrics = dict(
            train_accuracy=torchmetrics.Accuracy(),
            # (weighted)
            train_precision=torchmetrics.Precision(num_classes=3, average="weighted"),
            train_recall=torchmetrics.Recall(num_classes=3, average="weighted"),
            train_f1=torchmetrics.F1Score(num_classes=3, average="weighted"),
            # (micro)
            train_precision_micro=torchmetrics.Precision(
                num_classes=3, average="micro"
            ),
            train_recall_micro=torchmetrics.Recall(num_classes=3, average="micro"),
            train_f1_micro=torchmetrics.F1Score(num_classes=3, average="micro"),
            # (macro)
            train_precision_macro=torchmetrics.Precision(
                num_classes=3, average="macro"
            ),
            train_recall_macro=torchmetrics.Recall(num_classes=3, average="macro"),
            train_f1_macro=torchmetrics.F1Score(num_classes=3, average="macro"),
            # (per class)
            train_precision_class=torchmetrics.Precision(num_classes=3, average="none"),
            train_recall_class=torchmetrics.Recall(num_classes=3, average="none"),
            train_f1_class=torchmetrics.F1Score(num_classes=3, average="none"),
            # Validation metrics
            val_accuracy=torchmetrics.Accuracy(),
            # (weighted)
            val_precision=torchmetrics.Precision(num_classes=3, average="weighted"),
            val_recall=torchmetrics.Recall(num_classes=3, average="weighted"),
            val_f1=torchmetrics.F1Score(num_classes=3, average="weighted"),
            # (micro)
            val_precision_micro=torchmetrics.Precision(num_classes=3, average="micro"),
            val_recall_micro=torchmetrics.Recall(num_classes=3, average="micro"),
            val_f1_micro=torchmetrics.F1Score(num_classes=3, average="micro"),
            # (macro)
            val_precision_macro=torchmetrics.Precision(num_classes=3, average="macro"),
            val_recall_macro=torchmetrics.Recall(num_classes=3, average="macro"),
            val_f1_macro=torchmetrics.F1Score(num_classes=3, average="macro"),
            # (per class)
            val_precision_class=torchmetrics.Precision(num_classes=3, average="none"),
            val_recall_class=torchmetrics.Recall(num_classes=3, average="none"),
            val_f1_class=torchmetrics.F1Score(num_classes=3, average="none"),
            # Test metrics
            test_accuracy=torchmetrics.Accuracy(),
            # (weighted)
            test_precision=torchmetrics.Precision(num_classes=3, average="weighted"),
            test_recall=torchmetrics.Recall(num_classes=3, average="weighted"),
            test_f1=torchmetrics.F1Score(num_classes=3, average="weighted"),
            # (micro)
            test_precision_micro=torchmetrics.Precision(num_classes=3, average="micro"),
            test_recall_micro=torchmetrics.Recall(num_classes=3, average="micro"),
            test_f1_micro=torchmetrics.F1Score(num_classes=3, average="micro"),
            # (macro)
            test_precision_macro=torchmetrics.Precision(num_classes=3, average="macro"),
            test_recall_macro=torchmetrics.Recall(num_classes=3, average="macro"),
            test_f1_macro=torchmetrics.F1Score(num_classes=3, average="macro"),
            # (per class)
            test_precision_class=torchmetrics.Precision(num_classes=3, average="none"),
            test_recall_class=torchmetrics.Recall(num_classes=3, average="none"),
            test_f1_class=torchmetrics.F1Score(num_classes=3, average="none"),
        )

    def log_metrics(self, preds, y, type):
        for metric_name, metric in self.metrics.items():
            if metric_name.startswith(type):
                metric(preds.cpu(), y.cpu())
                value = metric.compute()
                if len(value.shape) > 0:
                    for idx, val in enumerate(value):
                        self.log(
                            f"{metric_name}_{idx}", val, on_epoch=True, on_step=False
                        )
                else:
                    self.log(metric_name, value, on_epoch=True, on_step=False)



import logging
import os
from pathlib import Path
from dataclasses import asdict
from typing import Any, Dict, List, Tuple
import inspect
import time
import warnings
from collections import defaultdict
import pytorch_lightning as pl
import numpy as np
import torch
from torch.utils.data import DataLoader

from transformers import (
    AdamW,
    AutoConfig,
    AutoModel,
    AutoModelForPreTraining,
    AutoModelForQuestionAnswering,
    AutoModelForSeq2SeqLM,
    AutoModelForSequenceClassification,
    AutoModelForTokenClassification,
    AutoModelWithLMHead,
    AutoTokenizer,
    PretrainedConfig,
    PreTrainedTokenizer,
    get_linear_schedule_with_warmup,
    MBartTokenizer,
)

from kogito.core.utils import (
    assert_all_frozen,
    lmap,
    flatten_list,
    pickle_save,
    save_json,
    freeze_params,
    calculate_rouge,
    ROUGE_KEYS,
    calculate_bleu_score,
)
from kogito.core.dataset import Seq2SeqDataset, MBartDataset
from kogito.models.bart.config import COMETBARTConfig
from kogito.core.callbacks import LoggingCallback

logger = logging.getLogger(__name__)
device = "cuda" if torch.cuda.is_available() else "cpu"


logger = logging.getLogger(__name__)


MODEL_MODES = {
    "base": AutoModel,
    "sequence-classification": AutoModelForSequenceClassification,
    "question-answering": AutoModelForQuestionAnswering,
    "pretraining": AutoModelForPreTraining,
    "token-classification": AutoModelForTokenClassification,
    "language-modeling": AutoModelWithLMHead,
    "summarization": AutoModelForSeq2SeqLM,
    "translation": AutoModelForSeq2SeqLM,
}


def use_task_specific_params(model, task):
    """Update config with summarization specific params."""
    task_specific_params = model.config.task_specific_params

    if task_specific_params is not None:
        pars = task_specific_params.get(task, {})
        logger.info(f"using task specific params for {task}: {pars}")
        model.config.update(pars)


class BaseTransformer(pl.LightningModule):
    def __init__(
        self,
        config: COMETBARTConfig,
        num_labels=None,
        mode="base",
        pretrained_config=None,
        tokenizer=None,
        model=None,
        **config_kwargs,
    ):
        """Initialize a model, tokenizer and pretrained_config."""
        super().__init__()
        self.config = config
        self.step_count = 0
        self.tfmr_ckpts = {}
        self.output_dir = Path(self.config.output_dir)
        cache_dir = self.config.cache_dir if self.config.cache_dir else None
        if pretrained_config is None:
            self.pretrained_config = AutoConfig.from_pretrained(
                self.config.pretrained_config
                if self.config.pretrained_config
                else self.config.pretrained_model,
                **({"num_labels": num_labels} if num_labels is not None else {}),
                cache_dir=cache_dir,
                **config_kwargs,
            )
        else:
            self.pretrained_config: PretrainedConfig = pretrained_config
        if tokenizer is None:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.config.pretrained_tokenizer
                if self.config.pretrained_tokenizer
                else self.config.pretrained_model,
                cache_dir=cache_dir,
            )
        else:
            self.tokenizer: PreTrainedTokenizer = tokenizer
        self.model_type = MODEL_MODES[mode]
        if model is None:
            self.model = self.model_type.from_pretrained(
                self.config.pretrained_model,
                from_tf=bool(".ckpt" in self.config.pretrained_model),
                config=self.pretrained_config,
                cache_dir=cache_dir,
            )
        else:
            self.model = model

    def load_hf_checkpoint(self, *args, **kwargs):
        self.model = self.model_type.from_pretrained(*args, **kwargs)

    def configure_optimizers(self):
        "Prepare optimizer and schedule (linear warmup and decay)"
        model = self.model
        no_decay = ["bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [
                    p
                    for n, p in model.named_parameters()
                    if not any(nd in n for nd in no_decay)
                ],
                "weight_decay": self.config.weight_decay,
            },
            {
                "params": [
                    p
                    for n, p in model.named_parameters()
                    if any(nd in n for nd in no_decay)
                ],
                "weight_decay": 0.0,
            },
        ]
        optimizer = AdamW(
            optimizer_grouped_parameters,
            lr=self.config.learning_rate,
            eps=self.config.adam_epsilon,
        )
        self.opt = optimizer

        scheduler = get_linear_schedule_with_warmup(
            self.opt,
            num_warmup_steps=self.config.warmup_steps,
            num_training_steps=self.total_steps,
        )
        scheduler = {"scheduler": scheduler, "interval": "step", "frequency": 1}
        return [optimizer], [scheduler]

    def test_step(self, batch, batch_nb):
        return self.validation_step(batch, batch_nb)

    def test_epoch_end(self, outputs):
        return self.validation_end(outputs)

    def setup(self, stage=None):
        train_batch_size = self.config.train_batch_size
        dataloader = self.get_dataloader("train", train_batch_size)
        self.train_loader = dataloader
        self.total_steps = (
            (len(dataloader.dataset) // (train_batch_size * max(1, self.config.gpus)))
            // self.config.accumulate_grad_batches
            * float(self.config.max_epochs)
        )

    def train_dataloader(self):
        return self.train_loader

    def val_dataloader(self):
        return self.get_dataloader("dev", self.config.eval_batch_size)

    def test_dataloader(self):
        return self.get_dataloader("test", self.config.eval_batch_size)

    def _feature_file(self, mode):
        return os.path.join(
            self.config.data_dir,
            "cached_{}_{}_{}".format(
                mode,
                list(filter(None, self.config.pretrained_model.split("/"))).pop(),
                str(self.config.max_seq_length),
            ),
        )

    @pl.utilities.rank_zero_only
    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        save_path = self.output_dir.joinpath("best_tfmr")
        save_path.mkdir(exist_ok=True)
        self.model.config.save_step = self.step_count
        self.model.save_pretrained(save_path)
        self.tokenizer.save_pretrained(save_path)
        self.tfmr_ckpts[self.step_count] = save_path


def generic_train(
    model: BaseTransformer,
    config: COMETBARTConfig,
    logger=True,  # can pass WandbLogger() here
    extra_callbacks=[],
    checkpoint_callback=None,
    logging_callback=None,
    **extra_train_kwargs,
):
    pl.seed_everything(config.seed)

    # init model
    odir = Path(model.config.output_dir)
    odir.mkdir(exist_ok=True)

    # add custom checkpoints
    if checkpoint_callback is None:
        checkpoint_callback = pl.callbacks.ModelCheckpoint(
            filepath=config.output_dir,
            prefix="checkpoint",
            monitor="val_loss",
            mode="min",
            save_top_k=1,
        )
    if logging_callback is None:
        logging_callback = LoggingCallback()

    train_params = {}

    # TODO: remove with PyTorch 1.6 since pl uses native amp
    if config.fp16:
        train_params["precision"] = 16
        train_params["amp_level"] = config.fp16_opt_level

    if config.gpus > 1:
        train_params["distributed_backend"] = "ddp"

    trainer_param_keys = inspect.signature(pl.Trainer).parameters.keys()

    trainer = pl.Trainer(
        **{
            ckey: cval
            for ckey, cval in asdict(config).items()
            if ckey in trainer_param_keys
        },
        weights_summary=None,
        callbacks=[logging_callback] + extra_callbacks,
        logger=logger,
        checkpoint_callback=checkpoint_callback,
        **train_params,
    )

    trainer.fit(model)

    return trainer


class SummarizationModule(BaseTransformer):
    mode = "summarization"
    loss_names = ["loss"]
    metric_names = ROUGE_KEYS
    val_metric = "rouge2"

    def __init__(self, config, train_graph, val_graph, test_graph, **kwargs):
        super().__init__(config, num_labels=None, mode=self.mode, **kwargs)
        self.train_graph = train_graph
        self.val_graph = val_graph
        self.test_graph = test_graph
        self.metrics_save_path = Path(self.output_dir) / "metrics.json"
        self.config_save_path = Path(self.output_dir) / "config.pkl"
        pickle_save(self.config, self.config_save_path)
        self.step_count = 0
        self.metrics = defaultdict(list)

        self.dataset_kwargs: dict = dict(
            train_graph=train_graph,
            val_graph=val_graph,
            test_graph=test_graph,
            max_source_length=self.config.max_source_length,
            prefix=self.model.config.prefix or "",
        )
        n_observations_per_split = {
            "train": self.config.n_train,
            "val": self.config.n_val,
            "test": self.config.n_test,
        }
        self.n_obs = {
            k: v if v >= 0 else None for k, v in n_observations_per_split.items()
        }

        self.target_lens = {
            "train": self.config.max_target_length,
            "val": self.config.val_max_target_length,
            "test": self.config.test_max_target_length,
        }
        assert (
            self.target_lens["train"] <= self.target_lens["val"]
        ), f"target_lens: {self.target_lens}"
        assert (
            self.target_lens["train"] <= self.target_lens["test"]
        ), f"target_lens: {self.target_lens}"

        if self.config.freeze_embeds:
            self.freeze_embeds()
        if self.config.freeze_encoder:
            freeze_params(self.model.get_encoder())
            assert_all_frozen(self.model.get_encoder())

        try:
            self.num_workers = config.num_workers
        except AttributeError:
            self.num_workers = 2

        self.decoder_start_token_id = None
        self.dataset_class = Seq2SeqDataset

    def freeze_embeds(self):
        try:
            freeze_params(self.model.model.shared)
            for d in [self.model.model.encoder, self.model.model.decoder]:
                freeze_params(d.embed_positions)
                freeze_params(d.embed_tokens)
        except AttributeError:
            freeze_params(self.model.shared)
            for d in [self.model.encoder, self.model.decoder]:
                freeze_params(d.embed_tokens)

    def forward(self, input_ids, **kwargs):
        return self.model(input_ids, **kwargs)

    def ids_to_clean_text(self, generated_ids: List[int]):
        gen_text = self.tokenizer.batch_decode(
            generated_ids, skip_special_tokens=True, clean_up_tokenization_spaces=True
        )
        return lmap(str.strip, gen_text)

    def _step(self, batch: dict) -> Tuple:
        pad_token_id = self.tokenizer.pad_token_id
        source_ids, source_mask, y = (
            batch["input_ids"],
            batch["attention_mask"],
            batch["decoder_input_ids"],
        )
        y_ids = y[:, :-1].contiguous()
        lm_labels = y[:, 1:].clone()
        lm_labels[y[:, 1:] == pad_token_id] = -100
        outputs = self(
            source_ids,
            attention_mask=source_mask,
            decoder_input_ids=y_ids,
            labels=lm_labels,
        )
        loss = outputs[0]
        return (loss,)

    def training_step(self, batch, batch_idx) -> Dict:
        loss_tensors = self._step(batch)
        logs = {name: loss for name, loss in zip(self.loss_names, loss_tensors)}
        self.log("train_loss", loss_tensors[0], on_epoch=True)
        return {"loss": loss_tensors[0], "log": logs}

    def validation_step(self, batch, batch_idx) -> Dict:
        result = self._generative_step(batch)
        self.log("val_loss", result["loss"], on_epoch=True)
        return result

    def validation_epoch_end(self, outputs, prefix="val") -> Dict:
        self.step_count += 1
        losses = {
            k: torch.stack([x[k] for x in outputs]).mean() for k in self.loss_names
        }
        loss = losses["loss"]
        rouges = {
            k: np.array([x[k] for x in outputs]).mean()
            for k in self.metric_names + ["gen_time", "summ_len"]
        }
        rouge_tensor: torch.FloatTensor = torch.tensor(rouges[self.val_metric]).type_as(
            loss
        )
        rouges.update({k: v.item() for k, v in losses.items()})
        losses.update(rouges)
        metrics = {f"{prefix}_avg_{k}": x for k, x in losses.items()}
        metrics["avg_rouge1"] = losses["rouge1"]
        metrics["step_count"] = self.step_count
        self.save_metrics(metrics, prefix)  # writes to self.metrics_save_path
        preds = flatten_list([x["preds"] for x in outputs])
        return {
            "log": metrics,
            "preds": preds,
            f"{prefix}_loss": loss,
            f"{prefix}_{self.val_metric}": rouge_tensor,
        }

    def save_metrics(self, latest_metrics, type_path) -> None:
        self.metrics[type_path].append(latest_metrics)
        save_json(self.metrics, self.metrics_save_path)

    def calc_generative_metrics(self, preds, target) -> Dict:
        return calculate_rouge(preds, target)

    def _generative_step(self, batch: dict) -> dict:
        pad_token_id = self.tokenizer.pad_token_id
        source_ids, source_mask, y = Seq2SeqDataset.trim_seq2seq_batch(
            batch, pad_token_id
        )
        t0 = time.time()
        generated_ids = self.model.generate(
            input_ids=source_ids,
            attention_mask=source_mask,
            use_cache=True,
            decoder_start_token_id=self.decoder_start_token_id,
        )
        gen_time = (time.time() - t0) / source_ids.shape[0]
        preds = self.ids_to_clean_text(generated_ids)
        target = self.ids_to_clean_text(y)
        loss_tensors = self._step(batch)
        base_metrics = {name: loss for name, loss in zip(self.loss_names, loss_tensors)}
        rouge: Dict = self.calc_generative_metrics(preds, target)
        summ_len = np.mean(lmap(len, generated_ids))
        base_metrics.update(
            gen_time=gen_time, summ_len=summ_len, preds=preds, target=target, **rouge
        )
        return base_metrics

    def test_step(self, batch, batch_idx):
        return self._generative_step(batch)

    def test_epoch_end(self, outputs):
        return self.validation_epoch_end(outputs, prefix="test")

    def get_dataset(self, type_path) -> Seq2SeqDataset:
        n_obs = self.n_obs[type_path]
        max_target_length = self.target_lens[type_path]
        dataset = self.dataset_class(
            self.tokenizer,
            type_path=type_path,
            n_obs=n_obs,
            max_target_length=max_target_length,
            **self.dataset_kwargs,
        )
        return dataset

    def get_dataloader(
        self, type_path: str, batch_size: int, shuffle: bool = False
    ) -> DataLoader:
        dataset = self.get_dataset(type_path)
        sampler = None
        if self.config.sortish_sampler and type_path == "train":
            assert self.config.gpus <= 1
            sampler = dataset.make_sortish_sampler(batch_size)
            shuffle = False

        dataloader = DataLoader(
            dataset,
            batch_size=batch_size,
            collate_fn=dataset.collate_fn,
            shuffle=shuffle,
            num_workers=self.num_workers,
            sampler=sampler,
        )
        return dataloader

    def train_dataloader(self) -> DataLoader:
        dataloader = self.get_dataloader(
            "train", batch_size=self.config.train_batch_size, shuffle=True
        )
        t_total = (
            (
                len(dataloader.dataset)
                // (self.config.train_batch_size * max(1, self.config.gpus))
            )
            // self.config.accumulate_grad_batches
            * float(self.config.max_epochs)
        )
        scheduler = get_linear_schedule_with_warmup(
            self.opt,
            num_warmup_steps=self.config.warmup_steps,
            num_training_steps=t_total,
        )
        if max(scheduler.get_last_lr()) > 0:
            warnings.warn("All learning rates are 0")
        self.lr_scheduler = scheduler
        return dataloader

    def val_dataloader(self) -> DataLoader:
        return self.get_dataloader("val", batch_size=self.config.eval_batch_size)

    def test_dataloader(self) -> DataLoader:
        return self.get_dataloader("test", batch_size=self.config.eval_batch_size)


class TranslationModule(SummarizationModule):
    mode = "translation"
    loss_names = ["loss"]
    metric_names = ["bleu"]
    val_metric = "bleu"

    def __init__(self, config, train_graph, val_graph, test_graph, **kwargs):
        super().__init__(config, train_graph, val_graph, test_graph, **kwargs)
        self.dataset_kwargs["src_lang"] = config.src_lang
        self.dataset_kwargs["tgt_lang"] = config.tgt_lang
        if self.model.config.decoder_start_token_id is None and isinstance(
            self.tokenizer, MBartTokenizer
        ):
            self.decoder_start_token_id = self.tokenizer.lang_code_to_id[
                config.tgt_lang
            ]
        if isinstance(self.tokenizer, MBartTokenizer):
            self.dataset_class = MBartDataset

    def calc_generative_metrics(self, preds, target) -> dict:
        return calculate_bleu_score(preds, target)



import torch
import numpy as np
import pandas as pd
from torch.utils.data import Dataset
from torch.optim import Adam
from torch import nn
import pytorch_lightning as pl
from transformers import (
    DistilBertTokenizer,
    DistilBertModel,
    PretrainedConfig,
    PreTrainedModel,
)

from kogito.core.processors.models.utils import Evaluator


class DistilBERTHeadDataset(Dataset):
    def __init__(self, data, tokenizer_type="uncased"):
        self.tokenizer = DistilBertTokenizer.from_pretrained(
            f"distilbert-base-{tokenizer_type}"
        )
        self.labels = (
            np.asarray(data["label"].to_list())
            if isinstance(data, pd.DataFrame)
            else None
        )
        texts = data["text"] if isinstance(data, pd.DataFrame) else data
        self.features = [
            self.tokenizer(
                text,
                padding="max_length",
                max_length=32,
                truncation=True,
                return_tensors="pt",
            )
            for text in texts
        ]

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):
        if self.labels is not None:
            return self.features[idx], self.labels[idx]
        return self.features[idx]


class DistilBERTConfig(PretrainedConfig):
    def __init__(
        self,
        num_classes=3,
        dropout=0.5,
        learning_rate=1e-4,
        freeze_emb=False,
        model_case="uncased",
        **kwargs,
    ):
        self.num_classes = num_classes
        self.dropout = dropout
        self.learning_rate = learning_rate
        self.freeze_emb = freeze_emb
        self.model_case = model_case
        super().__init__(**kwargs)


class DistilBERTClassifier(PreTrainedModel, Evaluator, pl.LightningModule):
    config_class = DistilBERTConfig

    def __init__(self, config: DistilBERTConfig):
        super().__init__(config)
        self.distilbert = DistilBertModel.from_pretrained(
            f"distilbert-base-{config.model_case}"
        )
        self.dropout = nn.Dropout(config.dropout)
        self.linear = nn.Linear(768, config.num_classes)

        if config.freeze_emb:
            for parameter in self.distilbert.parameters():
                parameter.requires_grad = False
            self.classifier = nn.Sequential(self.linear)
        else:
            self.classifier = nn.Sequential(self.dropout, self.linear)

        self.criterion = nn.BCEWithLogitsLoss()
        self.learning_rate = config.learning_rate

        self.save_hyperparameters(config.to_dict(), ignore="config")

    def forward(self, input_ids, mask):
        outputs = self.distilbert(
            input_ids=input_ids, attention_mask=mask, return_dict=False
        )
        outputs = self.classifier(outputs[0][:, 0, :])
        return outputs

    def training_step(self, batch, batch_idx):
        X, y = batch
        mask = X["attention_mask"]
        input_ids = X["input_ids"].squeeze(1)
        outputs = self.forward(input_ids, mask)
        train_loss = self.criterion(outputs, y.float())
        preds = torch.sigmoid(outputs)
        self.log("train_loss", train_loss, on_epoch=True)
        self.log_metrics(preds, y, type="train")
        return train_loss

    def validation_step(self, batch, batch_idx):
        X, y = batch
        mask = X["attention_mask"]
        input_ids = X["input_ids"].squeeze(1)
        outputs = self.forward(input_ids, mask)
        val_loss = self.criterion(outputs, y.float())
        preds = torch.sigmoid(outputs)
        self.log("val_loss", val_loss, on_epoch=True)
        self.log_metrics(preds, y, type="val")
        return val_loss

    def test_step(self, batch, batch_idx):
        X, y = batch
        mask = X["attention_mask"]
        input_ids = X["input_ids"].squeeze(1)
        outputs = self.forward(input_ids, mask)
        test_loss = self.criterion(outputs, y.float())
        preds = torch.sigmoid(outputs)
        self.log("test_loss", test_loss, on_epoch=True)
        self.log_metrics(preds, y, type="test")
        return test_loss

    def predict_step(self, batch, batch_idx):
        X = batch
        mask = X["attention_mask"]
        input_ids = X["input_ids"].squeeze(1)
        outputs = self.forward(input_ids, mask)
        preds = torch.sigmoid(outputs)
        return preds

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



from typing import Optional

import glob
import os
from pathlib import Path
from dataclasses import asdict
from tqdm import tqdm
import pytorch_lightning as pl
import torch

from transformers import AutoModelForSeq2SeqLM, AutoTokenizer

from kogito.core.knowledge import KnowledgeGraph
from kogito.core.relation import KG_RELATIONS
from kogito.core.utils import (
    pickle_save,
    chunks,
    trim_batch,
)
from kogito.core.callbacks import Seq2SeqLoggingCallback, get_checkpoint_callback
from kogito.core.model import KnowledgeModel
from kogito.models.bart.config import COMETBARTConfig
from kogito.models.bart.utils import (
    generic_train,
    SummarizationModule,
    TranslationModule,
)

device = "cuda" if torch.cuda.is_available() else "cpu"


class COMETBART(KnowledgeModel):
    """COMET knowledge model based on BART"""

    def __init__(self, config: COMETBARTConfig, **kwargs) -> None:
        """Initialize COMET model

        Args:
            config (COMETBARTConfig): Config to use
        """
        self.model = None
        self.tokenizer = None
        self.config = config
        self.kwargs = kwargs

    def train(
        self,
        train_graph: KnowledgeGraph,
        val_graph: KnowledgeGraph,
        test_graph: Optional[KnowledgeGraph] = None,
        logger_name: str = "default",
    ) -> KnowledgeModel:
        """Train a COMET model

        Args:
            train_graph (KnowledgeGraph): Training dataset
            val_graph (KnowledgeGraph): Validation dataset
            test_graph (KnowledgeGraph, optional): Test dataset. Defaults to None
            logger_name (str, optional): Logger name to use. Accepted values: ["wandb", "default"]
                                         Defaults to "default".

        Raises:
            ValueError: When config.task is not recognized

        Returns:
            KnowledgeModel: Trained knowledge model
        """
        Path(self.config.output_dir).mkdir(exist_ok=True, parents=True)

        if self.config.task == "summarization":
            self.model: SummarizationModule = SummarizationModule(
                self.config, train_graph, val_graph, test_graph, **self.kwargs
            )
        elif self.config.task == "translation":
            self.model: SummarizationModule = TranslationModule(
                self.config, train_graph, val_graph, test_graph, **self.kwargs
            )
        else:
            raise ValueError("Unrecognized task")

        if self.config.atomic:
            self.model.tokenizer.add_tokens(
                [str(relation) for relation in KG_RELATIONS]
            )
            self.model.model.resize_token_embeddings(len(self.model.tokenizer))

        if (
            logger_name == "default"
            or str(self.config.output_dir).startswith("/tmp")
            or str(self.config.output_dir).startswith("/var")
        ):
            logger = True  # don't pollute wandb logs unnecessarily

        elif logger_name == "wandb":
            from pytorch_lightning.loggers import WandbLogger

            logger = WandbLogger(
                name=self.model.output_dir.name, project=self.model.output_dir.name
            )

        elif logger_name == "wandb_shared":
            from pytorch_lightning.loggers import WandbLogger

            logger = WandbLogger(
                name=self.model.output_dir.name,
                project=f"hf_{self.model.output_dir.name}",
            )

        trainer: pl.Trainer = generic_train(
            self.model,
            self.config,
            logging_callback=Seq2SeqLoggingCallback(),
            checkpoint_callback=get_checkpoint_callback(
                self.config.output_dir, self.model.val_metric
            ),
            logger=logger,
        )
        pickle_save(self.model.config, self.model.output_dir / "config.pkl")

        self.model.config.test_checkpoint = ""
        checkpoints = list(
            sorted(
                glob.glob(
                    os.path.join(self.config.output_dir, "*.ckpt"), recursive=True
                )
            )
        )
        if checkpoints:
            self.model.config.test_checkpoint = checkpoints[-1]
            trainer.resume_from_checkpoint = checkpoints[-1]
        trainer.logger.log_hyperparams(asdict(self.model.config))

        if test_graph:
            trainer.test(self.model)

    def generate(
        self, input_graph: KnowledgeGraph, batch_size: int = 64, **kwargs
    ) -> KnowledgeGraph:
        """Generate inferences from the model

        Args:
            input_graph (KnowledgeGraph): Input dataset
            batch_size (int, optional): Batch size to use. Defaults to 64.
            kwargs: Additional arguments to pass to the model.generate() function

        Returns:
            KnowledgeGraph: Complete knowledge graph
        """
        with torch.no_grad():
            outputs = []
            for kg_batch in tqdm(list(chunks(input_graph, batch_size))):
                queries = []
                for kg_input in kg_batch:
                    queries.append(
                        "{} {} [GEN]".format(str(kg_input.head), str(kg_input.relation))
                    )
                batch = self.tokenizer(
                    queries, return_tensors="pt", truncation=True, padding="max_length"
                ).to(device)
                input_ids, attention_mask = trim_batch(
                    **batch, pad_token_id=self.tokenizer.pad_token_id
                )

                if "num_beams" not in kwargs:
                    kwargs["num_beams"] = 3

                if "num_return_sequences" not in kwargs:
                    kwargs["num_return_sequences"] = 3

                if "max_length" not in kwargs:
                    kwargs["max_length"] = 24

                if "min_length" not in kwargs:
                    kwargs["min_length"] = 1

                summaries = self.model.generate(
                    input_ids=input_ids,
                    attention_mask=attention_mask,
                    decoder_start_token_id=self.config.decoder_start_token_id,
                    **kwargs,
                )

                output = self.tokenizer.batch_decode(
                    summaries,
                    skip_special_tokens=True,
                    clean_up_tokenization_spaces=False,
                )

                for kg_input, generations in zip(
                    kg_batch, list(chunks(output, kwargs["num_return_sequences"]))
                ):
                    output_kg = kg_input.copy()
                    output_kg.tails = generations
                    outputs.append(output_kg)

            return KnowledgeGraph(outputs)

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str = "mismayil/comet-bart-ai2",
        task: str = "summarization",
    ) -> KnowledgeModel:
        """Load pretrained model

        Args:
            model_name_or_path (str, optional): HuggingFace model name or local model path.
                                                Defaults to "mismayil/comet-bart-ai2".
            task (str, optional): Task used in training. Defaults to "summarization".

        Returns:
            KnowledgeModel: Loaded knowledge model
        """
        config = COMETBARTConfig(task=task, decoder_start_token_id=None)
        comet_bart = cls(config)
        model = AutoModelForSeq2SeqLM.from_pretrained(model_name_or_path)
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        comet_bart.model = model
        comet_bart.tokenizer = tokenizer
        model.to(device)
        return comet_bart

    def save_pretrained(self, save_path: str) -> None:
        """Save pretrained model

        Args:
            save_path (str): Directory path to save model to
        """
        if hasattr(self.model, "model"):
            self.model.model.save_pretrained(save_path)
            self.model.tokenizer.save_pretrained(save_path)
        else:
            self.model.save_pretrained(save_path)
            self.tokenizer.save_pretrained(save_path)



from typing import Optional
import numpy as np
import torch
from torch import cuda
from torch.utils.data import DataLoader
from transformers import GPT2LMHeadModel, GPT2Tokenizer
from pytorch_lightning.loggers import WandbLogger
import pytorch_lightning as pl

from kogito.models.gpt2.utils import GPT2Finetuner
from kogito.core.dataset import KnowledgeDataset
from kogito.core.model import KnowledgeModel
from kogito.core.knowledge import KnowledgeGraph, GEN_TOKEN, EOS_TOKEN, PAD_TOKEN
from kogito.core.relation import KG_RELATIONS

device = "cuda" if cuda.is_available() else "cpu"


class COMETGPT2(KnowledgeModel):
    """COMET model based on GPT-2"""

    def __init__(self, model_name_or_path: str = "gpt2") -> None:
        """Initialize COMET model

        Args:
            model_name_or_path (str, optional): HuggingFace model name or local model path. Defaults to "gpt2".
        """
        self.model = GPT2LMHeadModel.from_pretrained(model_name_or_path)
        self.tokenizer = GPT2Tokenizer.from_pretrained(model_name_or_path)
        self.model.to(device)

    def train(
        self,
        train_graph: KnowledgeGraph,
        val_graph: KnowledgeGraph,
        batch_size: int = 8,
        in_len: int = 16,
        out_len: int = 34,
        summary_len: int = 0,
        epochs: int = 1,
        lr: float = 5e-5,
        seed: int = 42,
        log_wandb: bool = False,
        output_dir: Optional[str] = None,
    ) -> KnowledgeModel:
        """Train a COMET model

        Args:
            train_graph (KnowledgeGraph): Training dataset
            val_graph (KnowledgeGraph): Validation dataset
            batch_size (int, optional): Batch size. Defaults to 2.
            in_len (int, optional): Input length. Defaults to 16.
            out_len (int, optional): Output length. Defaults to 34.
            summary_len (int, optional): Summary length. Defaults to 0.
            epochs (int, optional): Number of epochs. Defaults to 3.
            lr (float, optional): Learning rate. Defaults to 1e-5.
            seed (int, optional): Random seed. Defaults to 42.
            log_wandb (bool, optional): Whether to log to wandb. Defaults to False.
            output_dir (Optional[str], optional): Directory to save intermediate model checkpoints. Defaults to None.

        Returns:
            KnowledgeModel: Trained knowledge model
        """
        torch.manual_seed(seed)
        np.random.seed(seed)
        torch.backends.cudnn.deterministic = True
        self.tokenizer.add_special_tokens(
            {
                "eos_token": EOS_TOKEN,
                "pad_token": PAD_TOKEN,
                "additional_special_tokens": [
                    str(relation) for relation in KG_RELATIONS
                ]
                + [GEN_TOKEN],
            }
        )

        train_dataset = KnowledgeDataset(
            train_graph,
            tokenizer=self.tokenizer,
            source_len=out_len,
            target_len=summary_len,
        )
        val_dataset = KnowledgeDataset(
            val_graph,
            tokenizer=self.tokenizer,
            source_len=in_len,
            target_len=out_len - in_len,
            is_eval=True,
        )

        self.model.resize_token_embeddings(len(self.tokenizer))

        train_params = {"batch_size": batch_size, "shuffle": True, "num_workers": 0}
        val_params = {"batch_size": batch_size, "shuffle": False, "num_workers": 0}

        train_loader = DataLoader(train_dataset, **train_params, drop_last=True)
        val_loader = DataLoader(val_dataset, **val_params, drop_last=True)

        logger = True

        if log_wandb:
            config = {
                "batch_size": batch_size,
                "epochs": epochs,
                "learning_rate": lr,
                "seed": seed,
                "in_len": in_len,
                "summary_len": summary_len,
                "out_len": out_len,
            }
            logger = WandbLogger(project="kogito-comet-gpt2")
            logger.experiment.config.update(config)

        finetuner = GPT2Finetuner(model=self.model, learning_rate=lr)
        trainer = pl.Trainer(
            default_root_dir=output_dir,
            max_epochs=epochs,
            logger=logger,
            accelerator="auto",
        )
        trainer.fit(
            finetuner, train_dataloaders=train_loader, val_dataloaders=val_loader
        )

        self.save_pretrained(f"{output_dir}/final_model")

        return self.model

    def generate(
        self, input_graph: KnowledgeGraph, max_input_length: int = 16, **kwargs
    ) -> KnowledgeGraph:
        """Generate inferences from knowledge model

        Args:
            input_graph (KnowledgeGraph): Input dataset
            max_input_length (int, optional): Maximum input length. Defaults to 16.
            kwargs: Additional arguments to pass to the model.generate() function

        Returns:
            KnowledgeGraph: Completed knowledge graph
        """
        params = {"batch_size": 1, "shuffle": False, "num_workers": 0}

        if "temperature" not in kwargs:
            kwargs["temperature"] = 0.7

        if "top_p" not in kwargs:
            kwargs["top_p"] = 0.9

        if "repetition_penalty" not in kwargs:
            kwargs["repetition_penalty"] = 1.2

        if "num_beams" not in kwargs:
            kwargs["num_beams"] = 1

        if "num_return_sequences" not in kwargs:
            kwargs["num_return_sequences"] = 1

        if "top_k" not in kwargs:
            kwargs["top_k"] = 1

        if "do_sample" not in kwargs:
            kwargs["do_sample"] = False

        if "max_length" not in kwargs:
            kwargs["max_length"] = 34

        dataset = KnowledgeDataset(
            input_graph,
            tokenizer=self.tokenizer,
            source_len=max_input_length,
            target_len=kwargs["max_length"] - max_input_length,
            is_eval=True,
        )
        loader = DataLoader(dataset, **params, drop_last=False)

        self.model.eval()

        outputs = []

        with torch.no_grad():
            for input_kg, data in zip(input_graph, loader):
                ids = data["source_ids"].to(device)
                mask = data["source_mask"].to(device)

                generated_ids = self.model.generate(
                    input_ids=ids, attention_mask=mask, **kwargs
                )

                generations = [
                    self.tokenizer.decode(g, clean_up_tokenization_spaces=True)
                    for g in generated_ids
                ]

                generations = [
                    g[g.find(GEN_TOKEN) + len(GEN_TOKEN) : g.find(EOS_TOKEN)].strip()
                    for g in generations
                ]

                output_kg = input_kg.copy()
                output_kg.tails = generations
                outputs.append(output_kg)

        return KnowledgeGraph(outputs)

    def save_pretrained(self, save_path: str) -> None:
        """Save pretrained model

        Args:
            save_path (str): Directory to save model to
        """
        if save_path:
            self.model.save_pretrained(save_path)
            self.tokenizer.save_pretrained(save_path)

    @classmethod
    def from_pretrained(
        cls, model_name_or_path: str = "mismayil/comet-gpt2-ai2"
    ) -> KnowledgeModel:
        """Load pretrained model

        Args:
            model_name_or_path (str, optional): HuggingFace model name or local model path.
                                                Defaults to "mismayil/comet-gpt2-ai2".

        Returns:
            KnowledgeModel: Loaded knowledge model
        """
        return cls(model_name_or_path)



import string
from abc import ABC, abstractmethod
from typing import List, Optional

from spacy.tokens import Doc
from spacy.language import Language
from spacy.lang.en.stop_words import STOP_WORDS

from kogito.core.head import KnowledgeHead, KnowledgeHeadType
from kogito.core.utils import IGNORE_WORDS


class KnowledgeHeadExtractor(ABC):
    """Base class for head extraction"""

    def __init__(self, name: str, lang: Optional[Language] = None) -> None:
        """Initialize a head extractor

        Args:
            name (str): Unique head extractor name
            lang (Optional[Language], optional): Spacy language pipeline to use. Defaults to None.
        """
        self.name = name
        self.lang = lang

    @abstractmethod
    def extract(self, text: str, doc: Optional[Doc] = None) -> List[KnowledgeHead]:
        """Extract heads from text

        Args:
            text (str): Text to extract from
            doc (Optional[Doc], optional): Spacy doc to use for extraction. Defaults to None.

        Raises:
            NotImplementedError: This method has to be implemented by
                                 concrete subclasses.

        Returns:
            List[KnowledgeHead]: List of extracted knowledge heads.
        """
        raise NotImplementedError


class SentenceHeadExtractor(KnowledgeHeadExtractor):
    """Extracts sentences as heads from text"""

    def extract(self, text: str, doc: Optional[Doc] = None) -> List[KnowledgeHead]:
        if not doc:
            doc = self.lang(text)

        heads = []

        for sentence in doc.sents:
            if sentence.text.strip():
                heads.append(
                    KnowledgeHead(
                        text=sentence.text,
                        type=KnowledgeHeadType.SENTENCE,
                        entity=sentence,
                    )
                )

        return heads


class NounPhraseHeadExtractor(KnowledgeHeadExtractor):
    """Extracts noun phrases as heads from text"""

    def extract(self, text: str, doc: Optional[Doc] = None) -> List[KnowledgeHead]:
        if not doc:
            doc = self.lang(text)

        heads = []
        head_texts = set()

        for token in doc:
            if (
                token.text.strip().lower() not in STOP_WORDS.union(IGNORE_WORDS)
                and token.pos_ == "NOUN"
            ):
                token_text = token.text.strip(string.punctuation + " ")
                if token_text not in head_texts and len(token_text) > 1:
                    head_texts.add(token_text)
                    heads.append(
                        KnowledgeHead(
                            text=token.text.strip(),
                            type=KnowledgeHeadType.NOUN_PHRASE,
                            entity=token,
                        )
                    )

        for phrase in doc.noun_chunks:
            clean_phrase = []
            phrase_doc = self.lang(phrase.text)

            for token in phrase_doc:
                if token.text.strip().lower() not in STOP_WORDS.union(IGNORE_WORDS):
                    clean_phrase.append(token.text)

            clean_text = " ".join(clean_phrase).strip(string.punctuation + " ")

            if clean_text and clean_text not in head_texts and len(clean_text) > 1:
                head_texts.add(clean_text)
                heads.append(
                    KnowledgeHead(
                        text=clean_text,
                        type=KnowledgeHeadType.NOUN_PHRASE,
                        entity=phrase,
                    )
                )

        return heads


class VerbPhraseHeadExtractor(KnowledgeHeadExtractor):
    """Extracts verb phrases as heads from text"""

    def extract(self, text: str, doc: Optional[Doc] = None) -> List[KnowledgeHead]:
        if not doc:
            doc = self.lang(text)

        heads = []
        head_texts = set()

        for token in doc:
            if token.pos_ == "VERB":
                verb_text = f"to {token.lemma_}"

                if verb_text not in head_texts:
                    head_texts.add(verb_text)
                    heads.append(
                        KnowledgeHead(
                            text=verb_text,
                            type=KnowledgeHeadType.VERB_PHRASE,
                            entity=token,
                        )
                    )

                for child in token.children:
                    if child.dep_ in ("attr", "dobj"):
                        child_text = f"{token.lemma_} {child.text}"
                        if child_text not in head_texts:
                            head_texts.add(child_text)
                            heads.append(
                                KnowledgeHead(
                                    text=child_text,
                                    type=KnowledgeHeadType.VERB_PHRASE,
                                    entity=[token, child],
                                )
                            )

        return heads



import torch
import pytorch_lightning as pl


class GPT2Finetuner(pl.LightningModule):
    def __init__(self, model, learning_rate=1e-5) -> None:
        super().__init__()
        self.model = model
        self.learning_rate = learning_rate

    def forward(self, input_ids, mask):
        return self.model(input_ids=input_ids, attention_mask=mask, labels=input_ids)

    def training_step(self, batch, batch_idx):
        X = batch
        ids = X["source_ids"]
        mask = X["source_mask"]
        outputs = self.model(input_ids=ids, attention_mask=mask, labels=ids)
        loss = outputs[0]
        self.log("train_loss", loss, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        X = batch
        ids = X["source_ids"]
        mask = X["source_mask"]
        outputs = self.model(input_ids=ids, attention_mask=mask, labels=ids)
        loss = outputs[0]
        self.log("val_loss", loss, on_epoch=True)
        return loss

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



import numpy as np
import pandas as pd
import torch
from torch.optim import Adam
from torch import nn
import pytorch_lightning as pl
import torch.nn.functional as F
import spacy
from torch.utils.data import Dataset
from torch.nn.utils.rnn import pad_sequence
from transformers import PretrainedConfig, PreTrainedModel

from kogito.core.processors.models.utils import Evaluator, text_to_embedding


class SWEMHeadDataset(Dataset):
    def __init__(
        self,
        data,
        vocab,
        embedding_matrix=None,
        apply_pooling=False,
        pooling="avg",
        lang=None,
    ):
        texts = data["text"] if isinstance(data, pd.DataFrame) else data
        labels = data["label"] if isinstance(data, pd.DataFrame) else None

        if not lang:
            lang = spacy.load("en_core_web_sm")
        self.texts = []

        if apply_pooling:
            # Apply pooling directly without padding
            self.labels = []
            self.features = []

            for index, text in enumerate(texts):
                embedding = text_to_embedding(
                    text, vocab=vocab, embedding_matrix=embedding_matrix, lang=lang
                )
                if embedding is not None:
                    self.features.append(embedding)
                    if labels is not None:
                        self.labels.append(labels[index])
                    self.texts.append(text)

            self.labels = np.asarray(self.labels)
        else:
            # Pad sequences
            self.texts = texts
            self.labels = np.asarray(labels.to_list()) if labels is not None else None
            self.features = pad_sequence(
                [
                    torch.tensor(
                        [vocab.get(token.text, 1) for token in lang(text)],
                        dtype=torch.int,
                    )
                    for text in texts
                ],
                batch_first=True,
            )

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):
        if self.labels is not None:
            return self.features[idx], self.labels[idx]
        return self.features[idx]


class MaxPool(nn.Module):
    def forward(self, X):
        values, _ = torch.max(X, dim=1)
        return values


class AvgPool(nn.Module):
    def forward(self, X):
        return torch.mean(X, dim=1)


class SWEMConfig(PretrainedConfig):
    def __init__(
        self,
        num_classes=3,
        pooling="avg",
        freeze_emb=False,
        learning_rate=1e-4,
        num_embeddings=400002,
        embedding_dim=100,
        **kwargs
    ):
        self.num_classes = num_classes
        self.pooling = pooling
        self.freeze_emb = freeze_emb
        self.learning_rate = learning_rate
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        super().__init__(**kwargs)


class SWEMClassifier(PreTrainedModel, Evaluator, pl.LightningModule):
    config_class = SWEMConfig

    def __init__(self, config: SWEMConfig):
        super().__init__(config)

        try:
            embedding_matrix = np.load(
                "data/embedding_matrix_glove_100d.npy", allow_pickle=True
            )
            self.embedding = nn.Embedding(
                num_embeddings=embedding_matrix.shape[0],
                embedding_dim=embedding_matrix.shape[1],
            ).from_pretrained(
                torch.tensor(embedding_matrix, dtype=torch.float32),
                freeze=config.freeze_emb,
            )
        except FileNotFoundError:
            self.embedding = nn.Embedding(
                num_embeddings=config.num_embeddings,
                embedding_dim=config.embedding_dim,
            )

        self.pool = MaxPool() if config.pooling == "max" else AvgPool()
        self.linear = nn.Linear(self.embedding.embedding_dim, config.num_classes)
        self.model = nn.Sequential(self.embedding, self.pool, self.linear)
        self.criterion = nn.BCEWithLogitsLoss()
        self.learning_rate = config.learning_rate
        self.save_hyperparameters(config.to_dict(), ignore="config")

    def forward(self, X):
        outputs = self.model(X)
        probs = F.sigmoid(outputs)
        return probs

    def training_step(self, batch, batch_idx):
        X, y = batch
        outputs = self.model(X)
        train_loss = self.criterion(outputs, y.float())
        preds = self.forward(X)
        self.log("train_loss", train_loss, on_epoch=True)
        self.log_metrics(preds, y, type="train")
        return train_loss

    def validation_step(self, batch, batch_idx):
        X, y = batch
        outputs = self.model(X)
        val_loss = self.criterion(outputs, y.float())
        preds = self.forward(X)
        self.log("val_loss", val_loss, on_epoch=True)
        self.log_metrics(preds, y, type="val")
        return val_loss

    def test_step(self, batch, batch_idx):
        X, y = batch
        outputs = self.model(X)
        test_loss = self.criterion(outputs, y.float())
        preds = self.forward(X)
        self.log("test_loss", test_loss, on_epoch=True)
        self.log_metrics(preds, y, type="test")
        return test_loss

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



#!/usr/bin/env python

# Python wrapper for METEOR implementation, by Xinlei Chen
# Acknowledge Michael Denkowski for the generous discussion and help


from nltk.translate.meteor_score import meteor_score
from nltk.tokenize import word_tokenize


class Meteor:
    def __init__(self):
        pass

    def compute_score(self, gts, res):
        assert gts.keys() == res.keys()
        imgIds = gts.keys()
        scores = []

        for i in imgIds:
            assert len(res[i]) == 1
            score = round(
                meteor_score(
                    [word_tokenize(s) for s in gts[i]], word_tokenize(res[i][0])
                ),
                4,
            )
            scores.append(score)

        return sum(scores) / len(scores), scores

    def method(self):
        return "METEOR"



from typing import Optional
from dataclasses import dataclass, asdict

BART_TASKS = ["summarization", "translation"]
FP16_OPT_LEVELS = ["O0", "O1", "O2", "O3"]


@dataclass
class COMETBARTConfig:
    output_dir: str = None
    fp16: bool = False
    fp16_opt_level: str = "O2"
    tpu_cores: Optional[int] = None
    gradient_clip_val: float = 1.0
    accumulate_grad_batches: int = 1
    seed: int = 42
    max_source_length: int = 48
    max_target_length: int = 24
    val_max_target_length: int = 24
    test_max_target_length: int = 24
    freeze_encoder: bool = False
    freeze_embeds: bool = False
    sortish_sampler: bool = True
    n_train: int = -1
    n_val: int = 500
    n_test: int = -1
    task: str = "summarization"
    src_lang: str = ""
    tgt_lang: str = ""
    atomic: bool = True
    pretrained_model: str = "facebook/bart-large"
    pretrained_config: str = None
    pretrained_tokenizer: str = None
    cache_dir: str = ""
    learning_rate: float = 5e-5
    weight_decay: float = 0.0
    adam_epsilon: float = 1e-8
    warmup_steps: int = 0
    num_workers: int = 2
    max_epochs: int = 3
    train_batch_size: int = 32
    eval_batch_size: int = 32
    gpus: int = 1
    decoder_start_token_id: Optional[int] = None

    def __dict__(self):
        return asdict(self)



from typing import Callable, Optional, Literal
from enum import Enum
from kogito.core.head import KnowledgeHeadType
from kogito.core.utils import vp_present_participle, article, posessive


class KnowledgeRelationType(Enum):
    """
    Represents a Knowledge relation type.
    """

    TRANSOMCS = "transomcs"
    ATOMIC = "atomic"
    CONCEPTNET = "conceptnet"

    def __repr__(self):
        return str(self.value)


class KnowledgeRelation:
    """
    Represents a concept of Knowledge Relation.
    """

    def __init__(
        self,
        text: str,
        type: KnowledgeRelationType = KnowledgeRelationType.ATOMIC,
        verbalizer: Optional[Callable] = None,
        prompt: Optional[str] = None,
    ) -> None:
        """Initialize a KnowledgeRelation

        Args:
            text (str): Relation text.
            type (KnowledgeRelationType, optional): Relation type. Defaults to KnowledgeRelationType.ATOMIC.
            verbalizer (Optional[Callable], optional): Function to convert relation to natural text. Defaults to None.
            prompt (Optional[str], optional): Instructive text used to prompt NLG models. Defaults to None.
        """
        self.text = text
        self.type = type
        self.verbalizer = verbalizer
        self.prompt = prompt

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, KnowledgeRelation)
            and self.text == other.text
            and self.type == other.type
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self.text, self.type))

    def verbalize(
        self,
        head: str,
        tail: Optional[str] = None,
        include_tail: bool = False,
        **kwargs,
    ) -> Optional[str]:
        """Convert knowledge relation into natural text.

        Args:
            head (str): Knowledge head to use.
            tail (Optional[str], optional): Knowledge tail to use if any. Defaults to None.
            include_tail (bool, optional): Whether to include tail. Defaults to False.

        Returns:
            Optional[str]: Verbalized relation.
        """
        if self.verbalizer:
            kwargs["tail"] = tail
            text = self.verbalizer(head, **kwargs).strip()
            if include_tail and tail:
                return f"{text} {tail}"
            return f"{text} "

    @classmethod
    def from_text(
        cls, text: str, type: KnowledgeRelationType = KnowledgeRelationType.ATOMIC
    ) -> "KnowledgeRelation":
        """Initialize relation from text.

        Args:
            text (str): Relation text.
            type (KnowledgeRelationType, optional): Type of relation to use. Defaults to KnowledgeRelationType.ATOMIC.

        Returns:
            KnowledgeRelation: An instance of KnowledgeRelation
        """
        for relation in KG_RELATIONS:
            if relation.text == text:
                return relation

        return cls(text, type=type)

    def __repr__(self) -> str:
        return str(self.text)

    def copy(self) -> "KnowledgeRelation":
        """Copy itself

        Returns:
            KnowledgeRelation: Copied knowledge relation
        """
        return KnowledgeRelation(
            text=self.text,
            type=self.type,
            verbalizer=self.verbalizer,
            prompt=self.prompt,
        )


# Verbalizers
def at_location_verbalizer(head: str, **kwargs):
    text = f"You are likely to find {article(head)} {head} in"
    tail = kwargs.get("tail")
    if tail is not None:
        return f"{text} {article(tail)}"
    return text


def capable_of_verbalizer(head: str, **kwargs):
    return f"{head} can "


def causes_verbalizer(head: str, **kwargs):
    return f"Sometimes {head} causes "


def causes_desire_verbalizer(head: str, **kwargs):
    return f"{head} would make you want to "


def created_by_verbalizer(head: str, **kwargs):
    return f"{head} is created by"


def defined_as_verbalizer(head: str, **kwargs):
    return f"{head} is defined as "


def desires_verbalizer(head: str, **kwargs):
    return f"{article(head)} {head} desires"


def has_a_verbalizer(head: str, **kwargs):
    return f"{head} {posessive(head)} "


def sub_event_verbalizer(head: str, **kwargs):
    return f"While {vp_present_participle(head)}, you would "


def has_prerequisite_verbalizer(head: str, **kwargs):
    return f"{vp_present_participle(head)} requires "


def has_property_verbalizer(head: str, **kwargs):
    return f"{head} is "


def hindered_by_verbalizer(head: str, **kwargs):
    return f"{head}. This would not happen if"


def inherits_from_verbalizer(head: str, **kwargs):
    return f"{head} inherits from"


def instance_of_verbalizer(head: str, **kwargs):
    return f"{head} is an instance of"


def is_a_verbalizer(head: str, **kwargs):
    text = f"{head} is"
    tail = kwargs.get("tail")
    if tail is not None:
        return f"{text} {article(tail)}"
    return text


def made_of_verbalizer(head: str, **kwargs):
    return f"{head} is made of"


def made_up_of_verbalizer(head: str, **kwargs):
    return f"{article(head)} {head} contains"


def motivated_by_verbalizer(head: str, **kwargs):
    return f"You would {head} because you are "


def not_capable_of_verbalizer(head: str, **kwargs):
    return f"{head} is not capable of"


def not_desires_verbalizer(head: str, **kwargs):
    return f"{article(head)} {head} does not desire"


def not_has_a_verbalizer(head: str, **kwargs):
    return f"{head} does not have a"


def not_has_property_verbalizer(head: str, **kwargs):
    return f"{head} is not"


def not_is_a_verbalizer(head: str, **kwargs):
    return f"{head} is not"


def not_made_of_verbalizer(head: str, **kwargs):
    return f"{head} is not made of"


def object_use_verbalizer(head: str, **kwargs):
    return f"{article(head)} {head} can be used for"


def part_of_verbalizer(head: str, **kwargs):
    text = f"{article(head).upper()} {head} is a part of"
    tail = kwargs.get("tail")
    if tail is not None:
        return f"{text} {article(tail)}"
    return text


def receives_action_verbalizer(head: str, **kwargs):
    return f"{head} can be "


def related_to_verbalizer(head: str, **kwargs):
    return f"{head} is related to "


def symbol_of_verbalizer(head: str, **kwargs):
    return f"{head} is a symbol of"


def used_for_verbalizer(head: str, **kwargs):
    return f"{article(head).upper()} {head} is for "


def is_after_verbalizer(head: str, **kwargs):
    return f"{head}. Before that, "


def is_before_verbalizer(head: str, **kwargs):
    return f"{head}. After that, "


def is_filled_by_verbalizer(head: str, **kwargs):
    return f"{head} is filled by"


def o_effect_verbalizer(head: str, **kwargs):
    return f"{head}. The effect on others will be"


def o_react_verbalizer(head: str, **kwargs):
    return f"{head}. As a result, others feel"


def o_want_verbalizer(head: str, **kwargs):
    return f"{head}. After, others will want to"


def x_attr_verbalizer(head: str, **kwargs):
    return f"{head}. PersonX is"


def x_effect_verbalizer(head: str, **kwargs):
    return f"{head}. The effect on PersonX will be"


def x_intent_verbalizer(head: str, **kwargs):
    return f"{head}. PersonX did this to"


def x_need_verbalizer(head: str, **kwargs):
    return f"{head}. Before, PersonX needs to"


def x_react_verbalizer(head: str, **kwargs):
    return f"{head}. PersonX will be"


def x_reason_verbalizer(head: str, **kwargs):
    return f"{head}. PersonX did this because"


def x_want_verbalizer(head: str, **kwargs):
    return f"{head}. After, PersonX will want to"


# Relations
AT_LOCATION = KnowledgeRelation("AtLocation", verbalizer=at_location_verbalizer)
CAPABLE_OF = KnowledgeRelation("CapableOf", verbalizer=capable_of_verbalizer)
CAUSES = KnowledgeRelation("Causes", verbalizer=causes_verbalizer)
CAUSES_DESIRE = KnowledgeRelation(
    "CausesDesire",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=causes_desire_verbalizer,
)
CREATED_BY = KnowledgeRelation(
    "CreatedBy", type=KnowledgeRelationType.CONCEPTNET, verbalizer=created_by_verbalizer
)
DEFINED_AS = KnowledgeRelation(
    "DefinedAs", type=KnowledgeRelationType.CONCEPTNET, verbalizer=defined_as_verbalizer
)
DESIRE_OF = KnowledgeRelation("DesireOf", type=KnowledgeRelationType.CONCEPTNET)
DESIRES = KnowledgeRelation("Desires", verbalizer=desires_verbalizer)
HAS_A = KnowledgeRelation(
    "HasA", type=KnowledgeRelationType.CONCEPTNET, verbalizer=has_a_verbalizer
)
HAS_FIRST_SUB_EVENT = KnowledgeRelation(
    "HasFirstSubevent",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=sub_event_verbalizer,
)
HAS_LAST_SUB_EVENT = KnowledgeRelation(
    "HasLastSubevent",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=sub_event_verbalizer,
)
HAS_PAIN_CHARACTER = KnowledgeRelation(
    "HasPainCharacter", type=KnowledgeRelationType.CONCEPTNET
)
HAS_PAIN_INTENSITY = KnowledgeRelation(
    "HasPainIntensity", type=KnowledgeRelationType.CONCEPTNET
)
HAS_PREREQUISITE = KnowledgeRelation(
    "HasPrerequisite",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=has_prerequisite_verbalizer,
)
HAS_PROPERTY = KnowledgeRelation("HasProperty", verbalizer=has_property_verbalizer)
HAS_SUB_EVENT = KnowledgeRelation("HasSubEvent", verbalizer=sub_event_verbalizer)
HAS_SUB_EVENT2 = KnowledgeRelation(
    "HasSubevent",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=sub_event_verbalizer,
)
HINDERED_BY = KnowledgeRelation("HinderedBy", verbalizer=hindered_by_verbalizer)
INHERITS_FROM = KnowledgeRelation(
    "InheritsFrom",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=inherits_from_verbalizer,
)
INSTANCE_OF = KnowledgeRelation(
    "InstanceOf",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=instance_of_verbalizer,
)
IS_A = KnowledgeRelation(
    "IsA", type=KnowledgeRelationType.CONCEPTNET, verbalizer=is_a_verbalizer
)
LOCATED_NEAR = KnowledgeRelation("LocatedNear", type=KnowledgeRelationType.CONCEPTNET)
LOCATION_OF_ACTION = KnowledgeRelation(
    "LocationOfAction", type=KnowledgeRelationType.CONCEPTNET
)
MADE_OF = KnowledgeRelation(
    "MadeOf", type=KnowledgeRelationType.CONCEPTNET, verbalizer=made_of_verbalizer
)
MADE_UP_OF = KnowledgeRelation("MadeUpOf", verbalizer=made_up_of_verbalizer)
MOTIVATED_BY_GOAL = KnowledgeRelation(
    "MotivatedByGoal",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=motivated_by_verbalizer,
)
NOT_CAPABLE_OF = KnowledgeRelation(
    "NotCapableOf",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=not_capable_of_verbalizer,
)
NOT_DESIRES = KnowledgeRelation("NotDesires", verbalizer=not_desires_verbalizer)
NOT_HAS_A = KnowledgeRelation(
    "NotHasA", type=KnowledgeRelationType.CONCEPTNET, verbalizer=not_has_a_verbalizer
)
NOT_HAS_PROPERTY = KnowledgeRelation(
    "NotHasProperty",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=not_has_property_verbalizer,
)
NOT_IS_A = KnowledgeRelation(
    "NotIsA", type=KnowledgeRelationType.CONCEPTNET, verbalizer=not_is_a_verbalizer
)
NOT_MADE_OF = KnowledgeRelation(
    "NotMadeOf",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=not_made_of_verbalizer,
)
OBJECT_USE = KnowledgeRelation("ObjectUse", verbalizer=object_use_verbalizer)
PART_OF = KnowledgeRelation(
    "PartOf", type=KnowledgeRelationType.CONCEPTNET, verbalizer=part_of_verbalizer
)
RECEIVES_ACTION = KnowledgeRelation(
    "ReceivesAction",
    type=KnowledgeRelationType.CONCEPTNET,
    verbalizer=receives_action_verbalizer,
)
RELATED_TO = KnowledgeRelation(
    "RelatedTo", type=KnowledgeRelationType.CONCEPTNET, verbalizer=related_to_verbalizer
)
SYMBOL_OF = KnowledgeRelation(
    "SymbolOf", type=KnowledgeRelationType.CONCEPTNET, verbalizer=symbol_of_verbalizer
)
USED_FOR = KnowledgeRelation(
    "UsedFor", type=KnowledgeRelationType.CONCEPTNET, verbalizer=used_for_verbalizer
)
IS_AFTER = KnowledgeRelation("isAfter", verbalizer=is_after_verbalizer)
IS_BEFORE = KnowledgeRelation("isBefore", verbalizer=is_before_verbalizer)
IS_FILLED_BY = KnowledgeRelation("isFilledBy", verbalizer=is_filled_by_verbalizer)
O_EFFECT = KnowledgeRelation("oEffect", verbalizer=o_effect_verbalizer)
O_REACT = KnowledgeRelation("oReact", verbalizer=o_react_verbalizer)
O_WANT = KnowledgeRelation("oWant", verbalizer=o_want_verbalizer)
X_ATTR = KnowledgeRelation("xAttr", verbalizer=x_attr_verbalizer)
X_EFFECT = KnowledgeRelation("xEffect", verbalizer=x_effect_verbalizer)
X_INTENT = KnowledgeRelation("xIntent", verbalizer=x_intent_verbalizer)
X_NEED = KnowledgeRelation(
    "xNeed",
    verbalizer=x_need_verbalizer,
    prompt="What needs to be true for this event to take place?",
)
X_REACT = KnowledgeRelation("xReact", verbalizer=x_react_verbalizer)
X_REASON = KnowledgeRelation("xReason", verbalizer=x_reason_verbalizer)
X_WANT = KnowledgeRelation("xWant", verbalizer=x_want_verbalizer)

KG_RELATIONS = [
    AT_LOCATION,
    CAPABLE_OF,
    CAUSES,
    CAUSES_DESIRE,
    CREATED_BY,
    DEFINED_AS,
    DESIRE_OF,
    DESIRES,
    HAS_A,
    HAS_FIRST_SUB_EVENT,
    HAS_LAST_SUB_EVENT,
    HAS_PAIN_CHARACTER,
    HAS_PAIN_INTENSITY,
    HAS_PREREQUISITE,
    HAS_PROPERTY,
    HAS_SUB_EVENT,
    HAS_SUB_EVENT2,
    HINDERED_BY,
    INHERITS_FROM,
    INSTANCE_OF,
    IS_A,
    LOCATED_NEAR,
    LOCATION_OF_ACTION,
    MADE_OF,
    MADE_UP_OF,
    MOTIVATED_BY_GOAL,
    NOT_CAPABLE_OF,
    NOT_DESIRES,
    NOT_HAS_A,
    NOT_HAS_PROPERTY,
    NOT_IS_A,
    NOT_MADE_OF,
    OBJECT_USE,
    PART_OF,
    RECEIVES_ACTION,
    RELATED_TO,
    SYMBOL_OF,
    USED_FOR,
    IS_AFTER,
    IS_BEFORE,
    IS_FILLED_BY,
    O_EFFECT,
    O_REACT,
    O_WANT,
    X_ATTR,
    X_EFFECT,
    X_INTENT,
    X_NEED,
    X_REACT,
    X_REASON,
    X_WANT,
]

CONCEPTNET_RELATIONS = [
    relation
    for relation in KG_RELATIONS
    if relation.type == KnowledgeRelationType.CONCEPTNET
]

ATOMIC_RELATIONS = [
    relation
    for relation in KG_RELATIONS
    if relation.type == KnowledgeRelationType.ATOMIC
]

#: ATOMIC 2020 Physical relations
PHYSICAL_RELATIONS = [
    OBJECT_USE,
    CAPABLE_OF,
    MADE_UP_OF,
    HAS_PROPERTY,
    DESIRES,
    NOT_DESIRES,
    AT_LOCATION,
]

#: ATOMIC 2020 Event relations
EVENT_RELATIONS = [CAUSES, HINDERED_BY, X_REASON, IS_AFTER, IS_BEFORE, HAS_SUB_EVENT]

#: ATOMIC 2020 Social relations
SOCIAL_RELATIONS = [
    X_INTENT,
    X_REACT,
    O_REACT,
    X_ATTR,
    X_EFFECT,
    X_NEED,
    X_WANT,
    O_EFFECT,
    O_WANT,
]

NOUN_PHRASE_RELATIONS = PHYSICAL_RELATIONS
SENTENCE_RELATIONS = EVENT_RELATIONS + SOCIAL_RELATIONS
VERB_PHRASE_RELATIONS = EVENT_RELATIONS

HEAD_TO_RELATION_MAP = {
    KnowledgeHeadType.TEXT: SENTENCE_RELATIONS,
    KnowledgeHeadType.SENTENCE: SENTENCE_RELATIONS,
    KnowledgeHeadType.NOUN_PHRASE: NOUN_PHRASE_RELATIONS,
    KnowledgeHeadType.VERB_PHRASE: VERB_PHRASE_RELATIONS,
}

CONCEPTNET_TO_ATOMIC_MAP = {
    CAUSES: [CAUSES, X_EFFECT],
    CAUSES_DESIRE: X_WANT,
    MADE_OF: MADE_UP_OF,
    HAS_A: [MADE_UP_OF, HAS_PROPERTY],
    HAS_PREREQUISITE: X_NEED,
    HAS_SUB_EVENT2: HAS_SUB_EVENT,
    HAS_FIRST_SUB_EVENT: HAS_SUB_EVENT,
    HAS_LAST_SUB_EVENT: HAS_SUB_EVENT,
    MOTIVATED_BY_GOAL: [X_INTENT, X_REASON],
    PART_OF: MADE_UP_OF,
    USED_FOR: OBJECT_USE,
    RECEIVES_ACTION: [MADE_UP_OF, AT_LOCATION, CAUSES, OBJECT_USE],
}

CUSTOM_RELATIONS = []
RELATION_KIND = Literal["physical", "event", "social"]


def register_relation(
    relation: KnowledgeRelation, kind: Optional[RELATION_KIND] = None
) -> None:
    """Register a new relation based on its kind. By default, it registers the relation
    as a custom relation without a kind, but optionally it can be registered to be a Physical,
    Event or a Social relation.

    Args:
        relation (KnowledgeRelation): Knowledge relation to register.
        kind (Optional[RELATION_KIND], optional): Relation kind. Available kinds: physical, event, social.
                                                Defaults to None.
    """
    if relation not in KG_RELATIONS:
        KG_RELATIONS.append(relation)

        if kind == "physical":
            PHYSICAL_RELATIONS.append(relation)
        elif kind == "event":
            EVENT_RELATIONS.append(relation)
        elif kind == "social":
            SOCIAL_RELATIONS.append(relation)
        else:
            CUSTOM_RELATIONS.append(relation)


RELATION_TO_NL = {
    "AtLocation": "located or found at/in/on",
    "CapableOf": "is/are capable of",
    "Causes": "causes",
    "CausesDesire": "makes someone want",
    "CreatedBy": "is created by",
    "Desires": "desires",
    "HasA": "has, possesses or contains",
    "HasFirstSubevent": "begins with the event/action",
    "HasLastSubevent": "ends with the event/action",
    "HasPrerequisite": "to do this, one requires",
    "HasProperty": "can be characterized by being/having",
    "HasSubEvent": "includes the event/action",
    "HinderedBy": "can be hindered by",
    "InstanceOf": "is an example/instance of",
    "isAfter": "happens after",
    "isBefore": "happens before",
    "isFilledBy": "___ can be filled by",
    "MadeOf": "is made of",
    "MadeUpOf": "made (up) of",
    "MotivatedByGoal": "is a step towards accomplishing the goal",
    "NotDesires": "do(es) not desire",
    "ObjectUse": "used for",
    "UsedFor": "used for",
    "oEffect": "as a result, PersonY or others will",
    "oReact": "as a result, PersonY or others feels",
    "oWant": "as a result, PersonY or others wants",
    "PartOf": "is a part of",
    "ReceivesAction": "can receive or be affected by the action",
    "xAttr": "PersonX is seen as",
    "xEffect": "as a result, PersonX will",
    "xIntent": "because PersonX wants",
    "xNeed": "but before, PersonX needs",
    "xReact": "as a result, PersonX feels",
    "xReason": "because",
    "xWant": "as a result, PersonX wants",
}

RELATION_SIZE = {
    "ObjectUse": 165590,
    "AtLocation": 20221,
    "MadeUpOf": 3345,
    "HasProperty": 5617,
    "CapableOf": 7968,
    "Desires": 2737,
    "NotDesires": 2838,
    "isAfter": 22453,
    "HasSubEvent": 12845,
    "isBefore": 23208,
    "HinderedBy": 106658,
    "Causes": 376,
    "xReason": 334,
    "isFilledBy": 33266,
    "xNeed": 128955,
    "xAttr": 148194,
    "xEffect": 115124,
    "xReact": 81397,
    "xWant": 135360,
    "xIntent": 72677,
    "oEffect": 80166,
    "oReact": 67236,
    "oWant": 94548,
}



# Filename: cider.py
#
# Description: Describes the class to compute the CIDEr (Consensus-Based Image Description Evaluation) Metric
#               by Vedantam, Zitnick, and Parikh (http://arxiv.org/abs/1411.5726)
#
# Creation Date: Sun Feb  8 14:16:54 2015
#
# Authors: Ramakrishna Vedantam <vrama91@vt.edu> and Tsung-Yi Lin <tl483@cornell.edu>

from kogito.evaluation.cider.cider_scorer import CiderScorer


class Cider:
    """
    Main Class to compute the CIDEr metric

    """

    def __init__(self, test=None, refs=None, n=4, sigma=6.0):
        # set cider to sum over 1 to 4-grams
        self._n = n
        # set the standard deviation parameter for gaussian penalty
        self._sigma = sigma

    def compute_score(self, gts, res):
        """
        Main function to compute CIDEr score
        :param  hypo_for_image (dict) : dictionary with key <image> and value
                    <tokenized hypothesis / candidate sentence>
                ref_for_image (dict)  : dictionary with key <image> and value
                    <tokenized reference sentence>
        :return: cider (float) : computed CIDEr score for the corpus
        """

        assert gts.keys() == res.keys()
        imgIds = gts.keys()

        cider_scorer = CiderScorer(n=self._n, sigma=self._sigma)

        for id in imgIds:
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert type(hypo) is list
            assert len(hypo) == 1
            assert type(ref) is list
            assert len(ref) > 0

            cider_scorer += (hypo[0], ref)

        (score, scores) = cider_scorer.compute_score()

        return score, scores

    def method(self):
        return "CIDEr"



import logging
import os
from pathlib import Path

import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.utilities import rank_zero_only
from pytorch_lightning.utilities import rank_zero_info


def count_trainable_parameters(model):
    model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    params = sum([np.prod(p.size()) for p in model_parameters])
    return params


logger = logging.getLogger(__name__)


class Seq2SeqLoggingCallback(pl.Callback):
    @rank_zero_only
    def _write_logs(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        type_path: str,
        save_generations=True,
    ) -> None:
        logger.info(
            f"***** {type_path} results at step {trainer.global_step:05d} *****"
        )
        metrics = trainer.callback_metrics
        trainer.logger.log_metrics(
            {
                k: v
                for k, v in metrics.items()
                if k not in ["log", "progress_bar", "preds"]
            }
        )
        # Log results
        od = Path(pl_module.config.output_dir)
        if type_path == "test":
            results_file = od / "test_results.txt"
            generations_file = od / "test_generations.txt"
        else:
            # this never gets hit. I prefer not to save intermediate generations, and results are in metrics.json
            # If people want this it will be easy enough to add back.
            results_file = od / f"{type_path}_results/{trainer.global_step:05d}.txt"
            generations_file = (
                od / f"{type_path}_generations/{trainer.global_step:05d}.txt"
            )
            results_file.parent.mkdir(exist_ok=True)
            generations_file.parent.mkdir(exist_ok=True)
        with open(results_file, "a+") as writer:
            for key in sorted(metrics):
                if key in ["log", "progress_bar", "preds"]:
                    continue
                val = metrics[key]
                if isinstance(val, torch.Tensor):
                    val = val.item()
                msg = f"{key}: {val:.6f}\n"
                writer.write(msg)

        if not save_generations:
            return

        if "preds" in metrics:
            content = "\n".join(metrics["preds"])
            generations_file.open("w+").write(content)

    @rank_zero_only
    def on_train_start(self, trainer, pl_module):
        try:
            npars = pl_module.model.model.num_parameters()
        except AttributeError:
            npars = pl_module.model.num_parameters()

        n_trainable_pars = count_trainable_parameters(pl_module)
        # mp stands for million parameters
        trainer.logger.log_metrics(
            {"n_params": npars, "mp": npars / 1e6, "grad_mp": n_trainable_pars / 1e6}
        )

    @rank_zero_only
    def on_test_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        return self._write_logs(trainer, pl_module, "test")


def get_checkpoint_callback(output_dir, metric):
    """Saves the best model by validation ROUGE2 score."""
    if metric == "rouge2":
        exp = "{val_avg_rouge2:.4f}-{step_count}"
    elif metric == "bleu":
        exp = "{val_avg_bleu:.4f}-{step_count}"
    else:
        raise NotImplementedError(
            f"seq2seq callbacks only support rouge2 and bleu, got {metric},"
            "You can make your own by adding to this function."
        )

    checkpoint_callback = ModelCheckpoint(
        dirpath=output_dir,
        filename=exp,
        monitor=f"val_{metric}",
        mode="max",
        save_top_k=1,
    )
    return checkpoint_callback


class LoggingCallback(pl.Callback):
    def on_batch_end(self, trainer, pl_module):
        lrs = {f"lr_group_{i}": lr for i, lr in enumerate(self.lr_scheduler.get_lr())}
        pl_module.logger.log_metrics(lrs)

    def on_validation_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        rank_zero_info("***** Validation results *****")
        metrics = trainer.callback_metrics
        # Log results
        for key in sorted(metrics):
            if key not in ["log", "progress_bar"]:
                rank_zero_info("{} = {}\n".format(key, str(metrics[key])))

    def on_test_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        rank_zero_info("***** Test results *****")
        metrics = trainer.callback_metrics
        # Log and save results to file
        output_test_results_file = os.path.join(
            pl_module.base_config.output_dir, "test_results.txt"
        )
        with open(output_test_results_file, "w") as writer:
            for key in sorted(metrics):
                if key not in ["log", "progress_bar"]:
                    rank_zero_info("{} = {}\n".format(key, str(metrics[key])))
                    writer.write("{} = {}\n".format(key, str(metrics[key])))



#!/usr/bin/env python
#
# File Name : bleu.py
#
# Description : Wrapper for BLEU scorer.
#
# Creation Date : 06-01-2015
# Last Modified : Thu 19 Mar 2015 09:13:28 PM PDT
# Authors : Hao Fang <hfang@uw.edu> and Tsung-Yi Lin <tl483@cornell.edu>

from kogito.evaluation.bleu.bleu_scorer import BleuScorer


class Bleu:
    def __init__(self, n=4):
        # default compute Blue score up to 4
        self._n = n
        self._hypo_for_image = {}
        self.ref_for_image = {}

    def compute_score(self, gts, res):

        assert gts.keys() == res.keys()
        imgIds = gts.keys()

        bleu_scorer = BleuScorer(n=self._n)
        for id in imgIds:
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert type(hypo) is list
            assert len(hypo) == 1
            assert type(ref) is list
            assert len(ref) >= 1

            bleu_scorer += (hypo[0], ref)

        # score, scores = bleu_scorer.compute_score(option='shortest')
        score, scores = bleu_scorer.compute_score(option="closest", verbose=0)
        # score, scores = bleu_scorer.compute_score(option='average', verbose=0)

        # return (bleu, bleu_info)
        return score, scores

    def method(self):
        return "Bleu"



from typing import Union, List
import itertools
from transformers import DebertaV2ForSequenceClassification, DebertaV2Tokenizer
import spacy
import torch

from kogito.core.linker import KnowledgeLinker
from kogito.core.knowledge import KnowledgeGraph
from kogito.core.utils import truncate_sequences_dual, pad_ids
from kogito.core.relation import RELATION_TO_NL

NARRATIVE_SEP_TOKEN = "<n_sep>"
FACT_SEP_TOKEN = "<f_sep>"

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class DebertaLinker(KnowledgeLinker):
    def __init__(
        self,
        model_name_or_path: str = "mismayil/comfact-deberta-v2",
        language: str = "en_core_web_sm",
    ) -> None:
        super().__init__()
        self.tokenizer = DebertaV2Tokenizer.from_pretrained(model_name_or_path)
        self.model = DebertaV2ForSequenceClassification.from_pretrained(
            model_name_or_path
        )
        self.narrative_sep_id = self.tokenizer.convert_tokens_to_ids(
            NARRATIVE_SEP_TOKEN
        )
        self.fact_sep_id = self.tokenizer.convert_tokens_to_ids(FACT_SEP_TOKEN)
        self.pad_token_id = self.tokenizer.convert_tokens_to_ids(
            self.tokenizer.pad_token
        )
        self.nlp = spacy.load(language, exclude=["ner"])
        self.max_input_tokens = 512
        self.model.to(device)

    def save_pretrained(self, save_path: str) -> None:
        """Save pretrained model

        Args:
            save_path (str): Directory to save model to
        """
        if save_path:
            self.model.save_pretrained(save_path)
            self.tokenizer.save_pretrained(save_path)

    @classmethod
    def from_pretrained(cls, model_name_or_path: str) -> KnowledgeLinker:
        """Load pretrained linker

        Args:
            model_name_or_path (str): HuggingFace model name or local model path

        Returns:
            KnowledgeModel: Loaded knowledge linker
        """
        return cls(model_name_or_path)

    def link(
        self, input_graph: KnowledgeGraph, context: Union[List[str], str]
    ) -> List[float]:
        if isinstance(context, str):
            doc = self.nlp(context)
            sentences = []

            for sentence in doc.sents:
                sentences.append(sentence.text)

            context = sentences

        context_ids = [
            self.tokenizer.convert_tokens_to_ids(self.tokenizer.tokenize(sent))
            for sent in context
        ]

        input_ids = []

        for kg in input_graph:
            head = str(kg.head).strip().lower()
            relation = str(kg.relation).strip()

            if relation not in RELATION_TO_NL:
                raise ValueError(f"Invalid relation found: {relation}")

            relation = RELATION_TO_NL[relation].lower()

            head_ids = self.tokenizer.convert_tokens_to_ids(
                self.tokenizer.tokenize(head)
            )
            relation_ids = self.tokenizer.convert_tokens_to_ids(
                self.tokenizer.tokenize(relation)
            )

            for tail in kg.tails:
                tail = str(tail).strip().lower()
                tail_ids = self.tokenizer.convert_tokens_to_ids(
                    self.tokenizer.tokenize(tail)
                )
                fact_ids = [head_ids, relation_ids, tail_ids]
                truncated_context_ids = _truncate_context(
                    context_ids, fact_ids, self.max_input_tokens
                )
                context_ids_with_sep = list(
                    itertools.chain(
                        *[
                            ids + [self.narrative_sep_id]
                            for ids in truncated_context_ids[:-1]
                        ],
                        truncated_context_ids[-1],
                    )
                )
                fact_ids_with_sep = list(
                    itertools.chain(
                        *[ids + [self.fact_sep_id] for ids in fact_ids[:-1]],
                        fact_ids[-1],
                    )
                )
                input_ids.append(
                    self.tokenizer.build_inputs_with_special_tokens(
                        context_ids_with_sep, fact_ids_with_sep
                    )
                )

        input_ids = torch.tensor(pad_ids(input_ids, self.pad_token_id))

        with torch.no_grad():
            output = self.model(input_ids.to(device))
            unstacked_probs = torch.softmax(output.logits, dim=1)[:, 1].tolist()

        stacked_probs = []

        for kg in input_graph:
            probs = []
            for _ in range(len(kg.tails)):
                probs.append(unstacked_probs.pop(0))
            stacked_probs.append(probs)

        return stacked_probs


def _truncate_context(
    context_ids: List[List[int]], fact_ids: List[List[int]], max_tokens: int
) -> List[List[int]]:
    num_keep_tokens = (
        1 + len(context_ids) + len(fact_ids)
    )  # [CLS], [SEP], <d_sep> and <s_sep>

    for ids in fact_ids:
        num_keep_tokens += len(ids)  # do not truncate statement

    return truncate_sequences_dual(context_ids, max_tokens - num_keep_tokens)



import openai

from kogito.core.model import KnowledgeModel
from kogito.core.knowledge import KnowledgeGraph
from kogito.core.utils import get_uuid


class GPT3Zeroshot(KnowledgeModel):
    """Zeroshot knowledge model based on GPT-3"""

    def __init__(self, api_key: str, model_name: str = "text-davinci-002") -> None:
        """Initialize a GPT-3 model

        Args:
            api_key (str): OpenAI API Key for GPT-3 model
            model_name (str, optional): Type of GPT-3 model. Defaults to "text-davinci-002".
        """
        self.api_key = api_key
        self.model_name = model_name

    def train(self):
        raise ValueError("GPT-3 Zeroshot model is not trainable")

    def save_pretrained(self, save_model_path: str):
        raise ValueError("GPT-3 Zeroshot cannot be saved")

    @classmethod
    def from_pretrained(cls, model_name_or_path: str):
        raise ValueError("GPT-3 only supports API-based access")

    def generate(
        self,
        input_graph: KnowledgeGraph,
        num_samples: int = 10,
        include_task_prompt: bool = True,
        debug: bool = False,
        **kwargs,
    ) -> KnowledgeGraph:
        """Generate inferences from GPT-3 model

        Args:
            input_graph (KnowledgeGraph): Input dataset
            num_samples (int, optional): Number of samples to use. Defaults to 10.
            include_task_prompt (bool, optional): Whether to include task prompt. Defaults to True.
            debug (bool, optional): Whether to enable debug mode. Defaults to False.
            kwargs: Additional arguments to pass to the OpenAI.Completion API

        Returns:
            KnowledgeGraph: Completed knowledge graph
        """
        rel_kg_map = {}
        outputs = []

        if "max_tokens" not in kwargs:
            kwargs["max_tokens"] = 16

        if "temperature" not in kwargs:
            kwargs["temperature"] = 0.9

        if "top_p" not in kwargs:
            kwargs["top_p"] = 0.9

        if "n" not in kwargs:
            kwargs["n"] = 1

        if "logprobs" not in kwargs:
            kwargs["logprobs"] = None

        if "stop" not in kwargs:
            kwargs["stop"] = None

        for input_kg in input_graph:
            if input_kg.relation not in rel_kg_map:
                rel_kg_map[input_kg.relation] = {"samples": [], "targets": []}
            if input_kg.tails:
                rel_kg_map[input_kg.relation]["samples"].append(input_kg)
            else:
                rel_kg_map[input_kg.relation]["targets"].append(input_kg)

        for relation, kg_map in rel_kg_map.items():
            samples = kg_map["samples"][:num_samples]
            targets = kg_map["targets"]

            if targets:
                prompts = []
                sample_prompts = []

                for index, sample_kg in enumerate(samples):
                    sample_prompt = sample_kg.to_prompt(
                        include_tail=True, index=index + 1
                    )
                    sample_prompts.append(sample_prompt)

                final_sample_prompt = "\n\n".join(sample_prompts)

                for target in targets:
                    target_prompt = target.to_prompt(index=len(samples) + 1)
                    final_prompt = f"{final_sample_prompt}\n\n{target_prompt}"

                    if include_task_prompt and relation.prompt:
                        final_prompt = f"{relation.prompt}\n\n{final_prompt}"
                    prompts.append(final_prompt)

                response = complete_gpt3(
                    api_key=self.api_key,
                    model_name=self.model_name,
                    prompt=prompts,
                    max_tokens=kwargs["max_tokens"],
                    temperature=kwargs["temperature"],
                    top_p=kwargs["top_p"],
                    logprobs=kwargs["logprobs"],
                    n=kwargs["n"],
                    stop=kwargs["stop"],
                    debug=debug,
                )

                rel_outputs = []
                for target in targets:
                    rel_outputs.append(target.copy())

                for result in response.choices:
                    output_kg = rel_outputs[result["index"] // kwargs["n"]]
                    output_kg.tails.append(result["text"])

                outputs.extend(rel_outputs)

        return KnowledgeGraph(outputs)


def complete_gpt3(
    api_key,
    model_name,
    prompt,
    max_tokens=16,
    temperature=1,
    top_p=1,
    logprobs=None,
    n=1,
    stop=None,
    debug=False,
):
    response = None
    openai.api_key = api_key

    if debug:
        with open(f"gpt3_prompt_{get_uuid()}.txt", "w") as f:
            f.write("\n\n".join(prompt))

    try:
        response = openai.Completion.create(
            engine=model_name,
            prompt=prompt,
            max_tokens=max_tokens,
            temperature=temperature,
            logprobs=logprobs,
            top_p=top_p,
            echo=False,
            stop=stop,
            n=n,
        )
    except Exception as e:
        print("Something went wrong when querying GPT-3 API")
        raise e
    return response



from abc import ABC, abstractmethod
from typing import List, Tuple, Optional, Type
from functools import partial
import pkgutil
from io import BytesIO

import numpy as np
import torch
import pytorch_lightning as pl
from spacy.language import Language
from torch.utils.data import DataLoader, Dataset

from kogito.core.head import KnowledgeHead
from kogito.core.relation import (
    KnowledgeRelation,
    HEAD_TO_RELATION_MAP,
    PHYSICAL_RELATIONS,
    EVENT_RELATIONS,
    SOCIAL_RELATIONS,
    CUSTOM_RELATIONS,
)

from kogito.core.processors.models.swem import SWEMHeadDataset, SWEMClassifier
from kogito.core.processors.models.distilbert import (
    DistilBERTHeadDataset,
    DistilBERTClassifier,
)
from kogito.core.processors.models.bert import BERTHeadDataset, BERTClassifier

RELATION_CLASSES = [PHYSICAL_RELATIONS, EVENT_RELATIONS, SOCIAL_RELATIONS]


class KnowledgeRelationMatcher(ABC):
    """Base class for relation matching"""

    def __init__(self, name: str, lang: Optional[Language] = None) -> None:
        """Initialize relation matcher

        Args:
            name (str): Unique relation matcher name
            lang (Optional[Language], optional): Spacy language pipeline to use. Defaults to None.
        """
        self.name = name
        self.lang = lang

    @abstractmethod
    def match(
        self,
        heads: List[KnowledgeHead],
        relations: Optional[List[KnowledgeRelation]] = None,
        **kwargs
    ) -> List[Tuple[KnowledgeHead, KnowledgeRelation]]:
        """Match relations to given heads

        Args:
            heads (List[KnowledgeHead]): List of heads to match for.
            relations (Optional[List[KnowledgeRelation]], optional): Subset of relations to use for matching.
                                                                    Defaults to None.

        Raises:
            NotImplementedError: This method has to be implemented by
                                 concrete subclasses

        Returns:
            List[Tuple[KnowledgeHead, KnowledgeRelation]]: List of matched head, relation tuples
        """
        raise NotImplementedError


class BaseRelationMatcher(KnowledgeRelationMatcher):
    """Matches all relations with all heads"""

    def match(
        self,
        heads: List[KnowledgeHead],
        relations: List[KnowledgeRelation] = None,
        **kwargs
    ) -> List[Tuple[KnowledgeHead, KnowledgeRelation]]:
        head_relations = []

        for head in heads:
            rels_to_match = (
                PHYSICAL_RELATIONS
                + SOCIAL_RELATIONS
                + EVENT_RELATIONS
                + CUSTOM_RELATIONS
            )
            if relations:
                rels_to_match = set(rels_to_match).intersection(set(relations))
            for relation in rels_to_match:
                head_relations.append((head, relation))

        return head_relations


class SimpleRelationMatcher(KnowledgeRelationMatcher):
    """Matches relation based on simple heuristics"""

    def match(
        self,
        heads: List[KnowledgeHead],
        relations: List[KnowledgeRelation] = None,
        **kwargs
    ) -> List[Tuple[KnowledgeHead, KnowledgeRelation]]:
        head_relations = []

        for head in heads:
            rels_to_match = HEAD_TO_RELATION_MAP[head.type] + CUSTOM_RELATIONS
            if relations:
                rels_to_match = set(HEAD_TO_RELATION_MAP[head.type]).intersection(
                    set(relations)
                )
            for relation in rels_to_match:
                head_relations.append((head, relation))

        return head_relations


class ModelBasedRelationMatcher(KnowledgeRelationMatcher):
    """Matches relations based on relation classifiers"""

    def __init__(
        self,
        name: str,
        dataset_class: Type[Dataset],
        model_class: Type[pl.LightningModule],
        model_path: str,
        batch_size: int = 64,
        lang: Optional[Language] = None,
    ) -> None:
        """Initialize a model based relation matcher

        Args:
            name (str): Unique relation matcher name
            dataset_class (Type[Dataset]): Dataset class to use
            model_class (Type[pl.LightningModule]): Model class to use
            model_path (str): Model path to load model from
            batch_size (int, optional): Batch size for inference. Defaults to 64.
            lang (Optional[Language], optional): Spacy lang pipeline. Defaults to None.
        """
        super().__init__(name, lang)
        self.dataset_class = dataset_class
        self.model_class = model_class
        self.model_path = model_path
        self.batch_size = batch_size
        self.model = model_class.from_pretrained(model_path)

    def match(
        self,
        heads: List[KnowledgeHead],
        relations: List[KnowledgeRelation] = None,
        **kwargs
    ) -> List[Tuple[KnowledgeHead, KnowledgeRelation]]:
        data = [str(head) for head in heads]
        dataset = self.dataset_class(data)
        dataloader = DataLoader(dataset, batch_size=self.batch_size)
        trainer = pl.Trainer()
        predictions = torch.cat(
            trainer.predict(self.model, dataloaders=dataloader)
        ).numpy()
        head_relations = []

        for head, prob in zip(heads, predictions):
            prediction = np.where(prob >= 0.5, 1, 0).tolist()
            pred_rel_classes = [
                RELATION_CLASSES[idx]
                for idx, pred in enumerate(prediction)
                if pred == 1
            ]

            if not pred_rel_classes:
                pred_rel_classes = [RELATION_CLASSES[np.argmax(prob)]]

            for rel_class in pred_rel_classes:
                rels_to_match = rel_class
                if relations:
                    rels_to_match = set(rels_to_match).intersection(set(relations))
                for relation in rels_to_match:
                    head_relations.append((head, relation))

        return head_relations


class SWEMRelationMatcher(ModelBasedRelationMatcher):
    """Relation matcher based on Simple Word Embeddings (GloVe)"""

    def __init__(self, name: str, lang: Optional[Language] = None) -> None:
        vocab = np.load(
            BytesIO(pkgutil.get_data(__name__, "data/vocab_glove_100d.npy")),
            allow_pickle=True,
        ).item()
        dataset_class = partial(SWEMHeadDataset, vocab=vocab, lang=lang)
        model_class = SWEMClassifier
        model_path = "mismayil/kogito-rc-swem"
        super().__init__(
            name,
            dataset_class=dataset_class,
            model_class=model_class,
            model_path=model_path,
            lang=lang,
        )


class DistilBERTRelationMatcher(ModelBasedRelationMatcher):
    """Relation matcher based on DistilBERT embeddings"""

    def __init__(self, name: str, lang: Optional[Language] = None) -> None:
        dataset_class = DistilBERTHeadDataset
        model_class = DistilBERTClassifier
        model_path = "mismayil/kogito-rc-distilbert"
        super().__init__(
            name,
            dataset_class=dataset_class,
            model_class=model_class,
            model_path=model_path,
            lang=lang,
        )


class BERTRelationMatcher(ModelBasedRelationMatcher):
    """Relation matcher based on BERT embeddings"""

    def __init__(self, name: str, lang: Optional[Language] = None) -> None:
        dataset_class = BERTHeadDataset
        model_class = BERTClassifier
        model_path = "mismayil/kogito-rc-bert"
        super().__init__(
            name,
            dataset_class=dataset_class,
            model_class=model_class,
            model_path=model_path,
            lang=lang,
        )


class GraphBasedRelationMatcher(KnowledgeRelationMatcher):
    """Relation matcher based on knowledge graphs"""

    def match(
        self,
        heads: List[KnowledgeHead],
        relations: List[KnowledgeRelation] = None,
        **kwargs
    ) -> List[Tuple[KnowledgeHead, KnowledgeRelation]]:
        sample_graph = kwargs.get("sample_graph")
        head_relations = []

        if sample_graph:
            matched_rels = set()

            for kg in sample_graph:
                matched_rels.add(kg.relation)

            if relations:
                matched_rels = matched_rels.intersection(set(relations))

            for head in heads:
                for relation in matched_rels:
                    head_relations.append((head, relation))

        return head_relations
